################################################################################
#############                   Autocorrelation                   ##############
################################################################################

from tqdm.auto import tqdm
import numpy as np
import pandas as pd
from sklearn import preprocessing
import math

from ..aaindex import *

#list of amino acids
aminoAcids = [
    # "-",
    "A",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "V",
    "W",
    "Y",
]

def norm_moreaubroto_autocorrelation(
    sequence, lag=30, properties=["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102",
    "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]
    ):
    """
    Calculate Normalized MoreauBrotoAuto Autocorrelation (NMBAuto) descriptor.
    Autocorrelation descriptors are a class of topological descriptors,
    also known as molecular connectivity indices, that describe the level of
    correlation between two objects (protein or peptide sequences) in terms of
    their specific structural or physicochemical properties, which are
    defined based on the distribution of amino acid properties along the sequence.
    By default, 8 amino acid properties are used for deriving the descriptors. The derivations
    and detailed explanations of this type of descriptor is outlind in [2].
    The NMBAuto descriptor is a type of Autocorrelation descriptor that uses
    the property values as the basis for measurement. Each autocorrelation will
    generate the number of features depending on the lag value and number of
    properties input with total features = lag * number of properties. Using 
    the default 8 properties with default lag value of 30, 240 features are 
    generated, the default 8 properties are:

    AccNo. CIDH920105 - Normalized Average Hydrophobicity Scales
    AccNo. BHAR880101 - Average Flexibility Indices
    AccNo. CHAM820101 - Polarizability Parameter
    AccNo. CHAM820102 - Free Energy of Solution in Water, kcal/mole
    AccNo. CHOC760101 - Residue Accessible Surface Area in Tripeptide
    AccNo. BIGC670101 - Residue Volume
    AccNo. CHAM810101 - Steric Parameter
    AccNo. DAYM780201 - Relative Mutability

    Parameters
    ----------
    :sequence : str
        protein sequence in str form.
    :lag : int (default = 30)
        A value for a lag, the max value is equal to the length of shortest peptide minus one.

    Returns
    -------
    :norm_moreaubroto_autocorr_df : pd.Dataframe
        pandas Dataframe of NMBAuto values for protein sequence. Output will
        be of the shape N x 1, where N is the number of features calculated from
        the descriptor. By default, the shape will be 240 x 1 (30 features per 
        property - using 8 properties).

    References
    ----------
    [1]: D. Raimondi, G. Orlando, W. F. Vranken, and Y. Moreau,
        “Exploring the limitations of biophysical propensity scales coupled with
        machine learning for protein sequence analysis,” Sci. Rep., vol. 9, no. 1, p. 16932, 2019.
    [2]: S. A. K. Ong, H. H. Lin, Y. Z. Chen, Z. R. Li, and Z. Cao, “Efficacy
        of different protein descriptors in predicting protein functional families,”
        BMC Bioinformatics, vol. 8, p. 300, 2007.
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #set default lag if invalid value input
    if (lag>=len(sequence) or (lag<0) or not (isinstance(lag, int))):
        lag=30

    #validate at least 1 property input to function
    if (properties == "" or properties == []):
        raise ValueError('At least one property must be passed into function.')

    #if 1 property passed into function, wrap it to an iterable list
    if (isinstance(properties, str)):
        properties = [properties]
    
    #initialise dicts to store AAI properties and values
    aai_properties = {}
    aaindex = AAIndex()
    for prop in properties:
        aai_properties[prop] = {}

    #iterate through list of properties, getting property values from AAIndex
    for prop in properties:

        #get property values from AAIndex & reshape
        aaindex[prop]['values'].pop('-', None)
        prop_aminoacid_values = aaindex[prop]['values']
     
        #normalise property values, calculate mean and std dev
        aai_property_vals = {}
        for i, j in prop_aminoacid_values.items():
            aai_property_vals[i] = (j - (sum(prop_aminoacid_values.values()) / 
                len(prop_aminoacid_values.values()))) / _std(prop_aminoacid_values.values(), ddof=0)

        aa_counter = 0
        #assign property and associated amino acid values to aai_property_vals array
        for i, j in aai_property_vals.items():
            aai_property_vals[aminoAcids[aa_counter]] = aai_property_vals[i]
            aa_counter+=1
        aai_properties[prop] = aai_property_vals

    #store resultant autocorrelation values
    result = {}

    #iterate through list of properties, calculating autocorrelation values for the sequence, append to results dict
    for key in aai_properties:
        temp = 0
        for i in range(1,lag+1):
            for j in range(len(sequence)-i):
                temp = temp + aai_properties[key][sequence[j]] * aai_properties[key][sequence[j+1]]
            if len(sequence) - i == 0:
                result["MoreauBrotoAuto_" + key + "_"+str(i)] = round(
                    temp / (len(sequence)), 3
                )
            else:
                result["MoreauBrotoAuto_" + key + "_"+str(i)] = round(
                    temp / (len(sequence) - i), 3
                )

    #transform values and columns to DataFrame
    result_df = pd.DataFrame([list(result.values())], columns=list(result.keys()))

    return result_df

def moran_autocorrelation(
    sequence, lag=30, properties=["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102",
    "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]
    ):
    """
    **refer to NMBAuto docstring for autocorrelation description.
    Moran autocorrelation (MAuto) utilizes property deviations from the
    average values.

    Parameters
    ----------
    **refer to NMBAuto docstring for autocorrelation parameters.

    Returns
    -------
    :moran_autocorr_df : pd.DataFrame
        pandas Dataframe of MAuto values for protein sequence. Output will
        be of the shape N x 1, where N is the number of features calculated from
        the descriptor. By default, the shape will be 240 x 1 (30 features per 
        property - using 8 properties).

    References
    ----------
    [1]: B. Hollas, “An analysis of the autocorrelation descriptor for molecules,”
        J. Math. Chem., vol. 33, no. 2, pp. 91–101, 2003.
    [2]: S. A. K. Ong, H. H. Lin, Y. Z. Chen, Z. R. Li, and Z. Cao, “Efficacy
        of different protein descriptors in predicting protein functional families,”
        BMC Bioinformatics, vol. 8, p. 300, 2007.    
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #set default lag if invalid value input
    if (lag>=len(sequence) or (lag<0) or not (isinstance(lag, int))):
        lag=30

    #validate at least 1 property input to function
    if (properties == "" or properties == []):
        raise ValueError('At least one property must be passed into function.')

    #if 1 property passed into function, wrap it to an iterable list
    if (isinstance(properties, str)):
        properties = [properties]

    #initialise dicts to store AAI properties and values
    aai_properties = {}
    aaindex = AAIndex()
    for prop in properties:
        aai_properties[prop] = {}

    #iterate through list of properties, getting property values from AAIndex
    for prop in properties:
        #get property values from AAIndex & reshape
        aaindex[prop]['values'].pop('-', None)
        prop_aminoacid_values = aaindex[prop]['values']
        #normalise property values, calculate mean and std dev
        aai_property_vals = {}

        for i, j in prop_aminoacid_values.items():
            aai_property_vals[i] = (j - (sum(prop_aminoacid_values.values()) / 
                len(prop_aminoacid_values.values()))) / _std(prop_aminoacid_values.values(), ddof=0)

        aa_counter = 0
        #assign property and associated amino acid values to aai_property_vals array
        for i, j in aai_property_vals.items():
            aai_property_vals[aminoAcids[aa_counter]] = aai_property_vals[i]
            aa_counter+=1
        aai_properties[prop] = aai_property_vals

    #store resultant autocorrelation values
    result = {}

    #iterate through list of properties, calculating autocorrelation values for the sequence, append to results dict
    for key in aai_properties:
        cds = 0
        for aa in aminoAcids:
            cds = cds + sequence.count(aa) * aai_properties[key][aa]
        prop_mean = cds / len(sequence)

        cc = []
        for aa in sequence:
            cc.append(aai_properties[key][aa])

        # k = (np.std(cc, ddof=0)) **2
        k = (_std(cc,ddof=0)) ** 2

        for i in range(1,lag+1):
            temp = 0
            for j in range(len(sequence)-i):
                temp = temp + aai_properties[key][sequence[j]] - prop_mean * (
                    aai_properties[key][sequence[j+i]] - prop_mean
                )
            if len(sequence) - i == 0:
                result["MoranAuto_" + key + "_"+str(i)] = round(
                    temp / ((len(sequence)) / k, 5)
                )
            else:
                result["MoranAuto_" + key + "_"+str(i)] = round(
                    temp / ((len(sequence) - i)/k), 5
                )

    #transform values and columns to DataFrame
    result_df = pd.DataFrame([list(result.values())], columns=list(result.keys()))

    return result_df

def geary_autocorrelation(
    sequence, lag=30, properties=["CIDH920105", "BHAR880101", "CHAM820101", "CHAM820102",
    "CHOC760101", "BIGC670101", "CHAM810101", "DAYM780201"]
    ):
    """
    *refer to NMBAuto docstring for autocorrelation description.
    Geary Autocorrelation (GAuto) utilizes the square-difference of property
    values instead of vector-products (of property values or deviations).
    
    Parameters
    ----------
    **refer to NMBAuto docstring for autocorrelation parameters.

    Returns
    -------
    :geary_autocorr_df : pd.Series
        pandas Series of GAuto values for protein sequence. Output will
        be of the shape N x 1, where N is the number of features calculated from
        the descriptor. By default, the shape will be 240 x 1 (30 features per 
        property - using 8 properties).

    References
    ----------
    [1]: B. Hollas, “An analysis of the autocorrelation descriptor for molecules,”
        J. Math. Chem., vol. 33, no. 2, pp. 91–101, 2003.
    [2]: S. A. K. Ong, H. H. Lin, Y. Z. Chen, Z. R. Li, and Z. Cao, “Efficacy
        of different protein descriptors in predicting protein functional families,”
        BMC Bioinformatics, vol. 8, p. 300, 2007. 
    """
    #check input sequence is a string, if not raise type error
    if not isinstance(sequence, str):
        raise TypeError('Input sequence must be a string, got input of type {}'.format(type(sequence)))

    #set default lag if invalid value input
    if (lag>=len(sequence) or (lag<0) or not (isinstance(lag, int))):
        lag=30

    #validate at least 1 property input to function
    if (properties == "" or properties == []):
        raise ValueError('At least one property must be passed into function.')

    #if 1 property passed into function, wrap it to an iterable list
    if (isinstance(properties, str)):
        properties = [properties]
        
    #initialise dicts to store AAI properties and values
    aai_properties = {}
    aaindex = AAIndex()
    for prop in properties:
        aai_properties[prop] = {}

    #iterate through list of properties, getting property values from AAIndex
    for prop in properties:
        #get property values from AAIndex & reshape
        aaindex[prop]['values'].pop('-', None)
        prop_aminoacid_values = aaindex[prop]['values']
     
        #normalise property values, calculate mean and std dev
        aai_property_vals = {}
        for i, j in prop_aminoacid_values.items():
            aai_property_vals[i] = (j - (sum(prop_aminoacid_values.values()) / 
                len(prop_aminoacid_values.values()))) / _std(prop_aminoacid_values.values(), ddof=0)

        aa_counter = 0
        #assign property and associated amino acid values to aai_property_vals array
        for i, j in aai_property_vals.items():
            aai_property_vals[aminoAcids[aa_counter]] = aai_property_vals[i]
            aa_counter+=1
        aai_properties[prop] = aai_property_vals

    #store resultant autocorrelation values
    result = {}

    #iterate through list of properties, calculating autocorrelation values for the sequence, append to results dict
    for key in aai_properties:
        cc = []
        for aa in sequence:
            cc.append(aai_properties[key][aa])

        k = ((np.std(cc,ddof=0)) ** 2) * len(sequence) / (len(sequence) - 1)

        for i in range(1,lag+1):
            temp = 0
            for j in range(len(sequence)-i):
                temp = (temp + (
                    aai_properties[key][sequence[j]] - aai_properties[key][sequence[j+i]]) **2
                )
            if len(sequence) - i == 0:
                result["GearyAuto_" + key + "_"+str(i)] = round(
                    temp / (2* (len(sequence))) / k,3
                )
            else:
                result["GearyAuto_" + key + "_"+str(i)] = round(
                    temp / (2* (len(sequence) -i))/k,3
                )

    #transform values and columns to DataFrame
    result_df = pd.DataFrame([list(result.values())], columns=list(result.keys()))

    return result_df

def _std(array, ddof=1):
    """
    Calculate the standard deviation of the array data, with associated means Delta Degrees 
    of Freedom (ddof).

    Parameters
    ----------
    :array : np.array
        numpy array of floats.
    :ddof : int (default=1)
        Means Delta Degrees of Freedom.

    Returns
    -------
    :result : np.array
        input array after standard deviation transformation.
    """
    return math.sqrt(sum([math.pow(i - sum(array) / len(array), 2) for i in array]) 
        / (len(array) - ddof))



################################################################################
#################             Utilities Modules                #################
################################################################################

import pandas as pd
pd.options.mode.chained_assignment = None  #stop pandas warnings, default='warn'
import numpy as np
import os
import shutil
import csv

from .globals_ import DATA_DIR, OUTPUT_DIR, OUTPUT_FOLDER

def valid_sequence(sequences):
    """
    Function that iterates through all protein sequences and validates that
    each sequence is made up of valid canonical amino acid letters. If no
    invalid values are found then None will be returned. If invalid letters
    are found in the sequence, the sequence index and the index of the value
    in the sequence will be appened to a dict.

    Parameters
    ----------
    :sequences : list/np.ndarray
        list or array of protein sequences.

    Returns
    -------
    :None or invalid_indices : None/list
        if no invalid values found in the protein sequences, None returned. if
        invalid values found, list of dicts returned in the form
        {sequence index: invalid value in sequence index}.
    """
    #if input is string, cast to a list so it is iterable
    if isinstance(sequences,str):
        sequences = [sequences]

    #valid canonical amino acid letters
    valid_amino_acids = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M',\
        'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y','-']
    invalid_indices = []

    #iterate through all sequences, validating that there are no invalid values
    #present in the sequences, if there are then append to list of invalid indices
    for seq in range(0,len(sequences)):
        for aa in range(0,len(sequences[seq])):
            if (sequences[seq][aa] not in valid_amino_acids):
                invalid_indices.append({'Sequence #'+str(seq) : 'Index #'+str(aa)})

    #if no invalid values found in sequences return None, else return list of
    #dicts containing invalid index and invalid values
    if invalid_indices == []:
        return None
    else:
        return invalid_indices

def remove_gaps(sequences):
    """
    Function that removes any gaps ('-') from the protein sequences in the input.
    The descriptors cannot be calculated if a '-' value is passsed into their
    respective funtions so gaps need to be removed. Removing the gaps has the same
    effect as setting the value at the index of the sequence to 0 and has no effect
    on the descriptors calculation. Input can be a string, list of array of sequences.

    Parameters
    ----------
    :sequences : str/list/np.ndarray
        string of 1 protein sequence or array/list of protein sequences.

    Returns
    -------
    :protein_seqs : np.ndarray
        returns the same inputted protein sequences but with any gaps ('-') removed.
    """
    #bool needed to ensure correct output format if input is str
    is_string=False   

    #convert single string into 1 element list
    if isinstance(sequences, str):
      is_string = True
      sequences = [sequences]     

    #concatenate multiple sequences into 1 iterable list
    if isinstance(sequences, list) and \
      len(sequences)>1:
      # for i in range(0,len(protein_seqs)):
      #   protein_seqs[i] = ''.join(protein_seqs[i])
      sequences = [''.join(sequences)]

    #iterate through sequences, removing any gaps ('-')
    for row in range(0, len(sequences)):
        try:
            sequences[row] = sequences[row].replace("-","")
        except:
            raise ValueError('Error removing gaps from sequences at index {} '.format(row))

    #if input was str then join list of sequences into one str
    if is_string:
       sequences = ''.join(sequences)

    return sequences

def flatten(array):
    """
    Lambda function for flattening list of lists or array of lists into one
    1 Dimensional array/list. Input must contain an array of arrays of the same
    length. Input will be flattened into a 1-dimensional array of size M * N
    where M = len(array) and N = len(array[0]). The flattened output can then be
    reshaped into the required shape and format.

    Parameters
    ----------
    :array : np.ndarray / list
        array of arrays or list of lists to be flattened.

    Returns
    -------
    :flatten(array/list) : np.ndarray/list
        flattened 1 dimensional list or array.
    """
    #if input is a string then return input as cannot be flattened
    if isinstance(array, str):
        return array

    #create flatten lambda function
    flatten = lambda array: [item for sublist in array for item in sublist]

    #flatten array/list
    try:
        flattened_array = flatten(array)
    except:
        raise ValueError('Error flattening array of type: {} and size {} \
            '.format(type(array),len(array)))

    #if input is a numpy array then reshape to 1D numpy array else return list
    if isinstance(array,np.ndarray):
        return (np.array(flattened_array).reshape([-1,1]))
    else:
        return flattened_array

def zero_padding(sequences):
    """
    Pad sequences in input array with 0's such that every sequence is of the same length
    of max(len(sequences)).

    Parameters
    ----------
    :sequences : np.ndarray / list
        array or list of encoded protein sequences.

    Returns
    -------
    :sequences: np.ndarray / list
        input sequences but with every sequence in the object now zero paddded
        to be the same length.
    """
    #no need to zero-pad if only one sequence passed in
    if len(sequences) == 1:
        return sequences

    #get maximum length of all sequences
    max_len = len(max(sequences,key=len))

    #iterate through all sequences, padding with 0's to max_len
    for s in range(0,len(sequences)):
        if len(sequences[s])<max_len:
            sequences[s]+= str(0) * (max_len - len(sequences[s]))

    return sequences

def create_output_dir():
    """
    Create output directory pointed to by global OUTPUT_DIR folder and create a
    folder according to the OUTPUT_FOLDER global var within this directory,
    used for storing the outputs/results from current job. Each output folder will
    have a unique name as the current date & time (DateTime) will be used in its naming.

    """
    #if directory doesn't exist then create it
    if not os.path.isdir(OUTPUT_DIR):
        try:
            os.makedirs(OUTPUT_DIR)
        except:
            raise OSError('Error creating directory {} '.format(OUTPUT_DIR))

    #if output folder already exists then delete it
    if os.path.isdir(OUTPUT_FOLDER):
        shutil.rmtree(OUTPUT_FOLDER, ignore_errors=False, onerror=None)

    #create output folder in directory
    try:
        os.makedirs(OUTPUT_FOLDER)
    except:
        raise OSError('Error creating directory {} '.format(OUTPUT_FOLDER))

def save_results(results, name):
    """
    Save object DataFrame/Series containing metric names and their values captured from
    the encoding process. Save the results in this object to a CSV file named according
    to name input parameter. Function can also accept a dict of results.

    Parameters
    ----------
    :results : dict/pd.DataFrame/pd.Series
        object of the metrics and results from the encoding process. Ideally should
        be a dataframe/series but function also accepts a dict of results.
    :name : str
        name to call results file.
    """
    #output results to csv if results variable is a dictionary
    if isinstance(results,dict):
        with open(os.path.join(OUTPUT_FOLDER, name + '.csv'), 'w') as f:
            w = csv.DictWriter(f, results.keys())
            w.writeheader()
            w.writerow(results)

    #output results to csv if results variable is a dataframe or Series
    elif isinstance(results, pd.DataFrame) or isinstance(results,pd.Series):
        results.reset_index(drop=True, inplace=True)
        results.to_csv(os.path.join(OUTPUT_FOLDER, name+'.csv'))

    else:
        raise TypeError('Results Object must be of type: dict, pd.Series or pd.DataFrame, got object of type {}'
            .format(type(results)))



################################################################################
#################                    AAIndex                   #################
################################################################################

#importing required modules and dependencies
import json
import numpy as np
import pandas as pd
import os
import matplotlib.pyplot as plt
import numpy as np
from sklearn.preprocessing import OneHotEncoder, LabelEncoder
import sys, copy, re
import requests
import shutil
from difflib import get_close_matches
import csv
import urllib.request as request
from contextlib import closing
from collections import defaultdict
from difflib import get_close_matches
from .globals_ import DATA_DIR, OUTPUT_DIR, OUTPUT_FOLDER

class AAIndex():
    """
            Python parser for AAindex1: Amino Acid Index Database
                      (**abbreviated to AAI onwards**)

    The AAindex is a database of numerical indices representing various physiochemical
    and biochemical properties of amino acids and pairs of amino acids. This class 
    stores the amino acid index of 20 numerical values for the 20 amino acids 
    (http://www.genome.jp/aaindex/).

    Some of the parsing functionality is patrially inspired off the AAIndex1 parser
    by harmslab: https://github.com/harmslab/hops/blob/master/hops/features/data/util/aaindex2json.py

    Data format of AAI1:
    ************************************************************************
    *                                                                      *
    * Each entry has the following format.                                 *
    *                                                                      *
    * H Accession number                                                   *
    * D Data description                                                   *
    * R PMID                                                               *
    * A Author(s)                                                          *
    * T Title of the article                                               *
    * J Journal reference                                                  *
    * * Comment or missing                                                 *
    * C Accession numbers of similar entries with the correlation          *
    *   coefficients of 0.8 (-0.8) or more (less).                         *
    *   Notice: The correlation coefficient is calculated with zeros       *
    *   filled for missing values.                                         *
    * I Amino acid index data in the following order                       *
    *   Ala    Arg    Asn    Asp    Cys    Gln    Glu    Gly    His    Ile *
    *   Leu    Lys    Met    Phe    Pro    Ser    Thr    Trp    Tyr    Val *
    * //                                                                   *
    ************************************************************************
    --------------------------------------------------------------------------------
    H ANDN920101
    D alpha-CH chemical shifts (Andersen et al., 1992)
    R PMID:1575719
    A Andersen, N.H., Cao, B. and Chen, C.
    T Peptide/protein structure analysis using the chemical shift index method:
      upfield alpha-CH values reveal dynamic helices and aL sites
    J Biochem. and Biophys. Res. Comm. 184, 1008-1014 (1992)
    C BUNA790102    0.949
    I    A/L     R/K     N/M     D/F     C/P     Q/S     E/T     G/W     H/Y     I/V
        4.35    4.38    4.75    4.76    4.65    4.37    4.29    3.97    4.63    3.95
        4.17    4.36    4.52    4.66    4.44    4.50    4.35    4.70    4.60    3.95
    --------------------------------------------------------------------------------

    Attributes
    ----------
    :aaindex_filename : str (default = 'aaindex1')
        local filename of aaindex1 file that class will import the database
        from in the data directory, default value of 'aaindex1' is reccomended.
    :download_using : str (default = "ftp")
        decide to download AAI database using ftp or https, ftp used by default.

    Methods
    -------
    parse_aaindex():
        parse AAI database into JSON format.
    parse_categories(aaindex_category_file):
        parse AAI indices into their respective categories.
    download_aaindex():
        download AAI database from its FTP server.
    get_amino_acids():
        get list of 20 amino acid letters.
    get_amino_acids_encoding():
        get one-hot encoding of amino acids.
    record_codes():
        get list of AAI record codes/Accession numbers.
    get_num_records():
        return total number of records in AAI database.
    get_record_names():
        return list of all descriptions for all records in AAI database.
    get_record_from_desc():
        return AAI record from its description.
    get_values_from_record(code):
        return amino acid index values for AAI record usings its accession number.
    get_ref_from_record():
        return references for AAI record.
    get_category_from_record():
        return category of AAI index record.

    References
    ----------
    [1]: Nakai, K., Kidera, A., and Kanehisa, M.;  Cluster analysis of
        amino acid indices for prediction of protein structure and
        function.  Protein Eng. 2, 93-100 (1988)
    [2]: Tomii, K. and Kanehisa, M.;  Analysis of amino acid indices and
        mutation matrices for sequence comparison and structure
        prediction of proteins.  Protein Eng. 9, 27-36 (1996).
    [3]: Kawashima, S., Ogata, H., and Kanehisa, M.;  AAindex: amino acid
        index database.  Nucleic Acids Res. 27, 368-369 (1999).
    [4]: Kawashima, S. and Kanehisa, M.;  AAindex: amino acid index
        database.  Nucleic Acids Res. 28, 374 (2000).
    """
    def __init__(self, aaindex_filename='aaindex1', download_using='ftp'):

        self.aaindex_filename = aaindex_filename
        self.download_using = download_using
        self.pySAR_module_path = os.path.dirname(os.path.abspath(sys.modules[self.__module__].__file__))

        #download AAI database using ftp or https
        if self.download_using=='ftp':
            self.url = "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1"
        elif self.download_using=='http' or self.download_using=='https':
            self.url = "https://www.genome.jp/ftp/db/community/aaindex/aaindex1"
        else:
            self.url = "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1"

        #if AAIndex database not found in data directory then call download method
        if not (os.path.isfile(os.path.join(self.pySAR_module_path, DATA_DIR, self.aaindex_filename))):
            self.download_aaindex()

        #if parsed json of AAIndex already in file then read it and return 
        if (os.path.isfile(os.path.join(self.pySAR_module_path, DATA_DIR, self.aaindex_filename + '.json'))):
            with open(os.path.join(self.pySAR_module_path, DATA_DIR, self.aaindex_filename +'.json')) as aai_json:
                self.aaindex_json = json.load(aai_json)
        else:
            #parse AAIndex database file into JSON format
            self.aaindex_json = self.parse_aaindex()

        #get dict of categories
        self.categories = self.get_all_categories()

    def parse_aaindex(self):
        """
        Parse AAI database into JSON format. Each AAI record will be indexed by
        its feature code/index code, and will be in the format as shown in the
        docstring above. The file will be stored in a json file called according
        to (self.aaindex_filename) variable.

        Returns
        -------
        :aaindex_json : dict
          parsed AAI database in dict form.
        """
        #initialise keys of AAI database
        template_dict = {"H":[],
                         "D":[],
                         "R":[],
                         "A":[],
                         "*":[],
                         "T":[],
                         "J":[],
                         "C":[],
                         "I":[]}

        #open AAI file for reading and parsing, by default it should be stored in DATA_DIR
        try:
            tmp_filepath = os.path.join(self.pySAR_module_path, DATA_DIR, self.aaindex_filename)
            f = open(tmp_filepath,'r')
        except IOError:
            print('Error opening file, check filename = {} and is stored in {} directory.'.format(self.aaindex_filename, DATA_DIR))

        #read lines of file
        lines = f.readlines()
        f.close()

        #pattern that seperates each AAI record
        clean_up_pattern = re.compile("\"")

        #initilaise parsed AAI database dictionary
        aaindex_json = {}
        current_dict = copy.deepcopy(template_dict)

        '''
        iterate through each line in the AAI database, parsing each record and its
        amino acid values into its own entry in the dictionary. Each index/record
        is seperated by a '//'. Remove any duplicate records, set any missing
        ('-') or NA amino acid values to 0. Store resulting dict into aaindex_json
        instance variable.
        '''
        for l in lines:

            if l.startswith("//"):

                # deal with meta data
                name = " ".join(current_dict["H"])
                name = clean_up_pattern.sub("'",name)

                description = " ".join(current_dict["D"])
                description = clean_up_pattern.sub("'",description)

                citation = "{} '{}' {}".format(" ".join(current_dict["A"]),
                                               " ".join(current_dict["T"]),
                                               " ".join(current_dict["J"]))

                citation = citation + "; Kawashima, S. and Kanehisa, M. \
                    'AAindex: amino acid index database.'  Nucleic Acids Res. 28, 374 (2000)."

                citation = clean_up_pattern.sub("'",citation)

                notes = " ".join(current_dict["*"])
                notes = clean_up_pattern.sub("'",notes)

                # parse amino acid data
                aa_lines = current_dict["I"]

                aa_names = aa_lines[0].split()
                row_0_names = [aa.split("/")[0] for aa in aa_names]
                row_1_names = [aa.split("/")[1] for aa in aa_names]

                row_0_values = aa_lines[1].split()
                row_1_values = aa_lines[2].split()

                #get values for each row of amino acids in record
                values = {}
                for i in range(len(row_0_values)):
                    try:
                        values[row_0_names[i]] = float(row_0_values[i])
                    except ValueError:
                        values[row_0_names[i]] = "NA"

                    try:
                        values[row_1_names[i]] = float(row_1_values[i])
                    except ValueError:
                        values[row_1_names[i]] = "NA"

                #look for duplicate name entries
                try:
                    aaindex_json[name]
                    err = "duplicate value name ({})".format(name)
                    raise ValueError('Duplicate AAI Record name found')
                except KeyError:
                    pass

                aaindex_json[name] = {"description":description,
                                  "refs":citation,
                                  "notes":notes,
                                  "values":values}

                current_dict = copy.deepcopy(template_dict)
                continue

            this_entry = l[0]
            if l[0] != " ":
                current_entry = this_entry

            current_dict[current_entry].append(l[1:].strip())
        
        #get categories for each record
        categories = self.get_all_categories()

        #append '-' to each aa index entry to account for missing AA in a protein sequence
        for index in aaindex_json:

            #add category to aaindex file
            aaindex_json[index]['category'] = categories[index]
            aaindex_json[index]['values']['-'] = 0

            #set any NA amino acid values to 0
            for val in aaindex_json[index]['values']:
                if aaindex_json[index]['values'][val] == 'NA':
                    aaindex_json[index]['values'][val] = 0

        #save parsed dictionary into JSON format to DATA_DIR
        with open((os.path.join(self.pySAR_module_path, DATA_DIR, self.aaindex_filename + '.json')),'w') as output_F:
          json.dump(aaindex_json, output_F, indent=4, sort_keys=True)

        return aaindex_json

    def download_aaindex(self, save_dir=DATA_DIR):
        """
        If AAI database not found in DATA directory, then it will be downloaded from
        the dedicated FTP or HTTPS server from https://www.genome.jp/aaindex/.
        FTP is the default method used for downloading the database with the URL:
        "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1". If you want to
        download by https, set the 'download_using' instance variable to 'https'. pySAR
        looks for the AAI database and its parsed JSON form in the 'data' directory, an
        error will throw if not found.

        Parameters
        ----------
        :save_dir : str (default = DATA_DIR)
            Directory to save the AAI database to. By default it wil be stored in
            the global var DATA_DIR = 'data'. Parameter has to point to a directory
            within the pySAR directory or there will be an error on instantiation 
            of the class. 
        """
        #fetch AAI database from URL if not present in DATA_DIR
        try:
            if not(os.path.isfile(os.path.join(self.pySAR_module_path, save_dir, self.aaindex_filename))):
                try:
                    with closing(request.urlopen(self.url)) as r:
                        with open((os.path.join(self.pySAR_module_path, save_dir, self.aaindex_filename)), 'wb') as f:
                            shutil.copyfileobj(r, f)
                    print('AAIndex1 successfully downloaded.')
                except requests.exceptions.RequestException:
                    print('Error downloading or exporting AAIndex from the url {}.'.format(self.url))
            else:
                pass    #AAIndex already in folder
        except:
            raise OSError('Save Directory does not exist: {}.'.format(save_dir))

    def parse_categories(self, aaindex_category_file='aaindex_to_category.txt'):
        """
        Parse category file which maps each AAI record in the databse into 1 of 8 categories.
        Category file and parsing code inspired from: https://github.com/harmslab/hops. 

        Parameters
        ----------
        :aaindex_category_file : str
            Name of category file to parse (default is "aaindex_to_category.txt"). If parameter
            is a file name then it will be imported from the 'data' directory by default, else
            it will be imported from its full path.

        Returns
        -------
        :aaindex_category : dict
            Dictionary that maps each AAI record into 1 of 8 categories.
        """
        #if input parameter is a full path, read it else read from default DATA_DIR
        if os.path.isfile(aaindex_category_file):
            try:
                f = open(aaindex_category_file,'r')
            except IOError:
                print('Error opening AAIndex1 category file {} .'
                    .format(aaindex_category_file))
        else:
            try:
                f = open((os.path.join(self.pySAR_module_path, DATA_DIR, aaindex_category_file)),'r')
            except IOError:
                print('Error opening AAIndex1 category file, check {} in {} directory'
                    .format(aaindex_category_file, DATA_DIR))

        #get total number of lines in file
        # total_lines = len(f.readlines(  ))

        #open new file in data directory to store parsed category file
        f.seek(0)
        category_output_file = "aaindex_categories.txt"
        f_out = open((os.path.join(self.pySAR_module_path, DATA_DIR, category_output_file)), "w")

        #lines starting with '#' are file metadata so don't write these to parsed output file
        for line in f.readlines():
            if not (line.startswith('#')):
                f_out.write(line)
        #close both files
        f.close()
        f_out.close()

        #open parsed category file for reading
        with open(os.path.join(self.pySAR_module_path, DATA_DIR, category_output_file)) as f_out:
            reader = csv.reader(f_out, delimiter="\t")
            d = list(reader)

        aaindex_category = {}

        #iterate through all lines in parsed file and store AAI indices as keys
        #  in the output dict and their respective categories as the values of dict
        for i in range(0,len(d)):
          for j in range(0,len(d[i])):
            category_substring = d[i][1].strip()
            category_substring = category_substring.split(" ",1)
            aaindex_category[d[i][0]] = category_substring[0]

        self.categories = aaindex_category

    def get_all_categories(self, category_file="aaindex_categories.txt"):
        """
        Return list of all indices, their description and associated categories from 
        parsed aaindex_categories.txt file created from parse_categories() function.

        Parameters
        ----------
        :category_file : str
            path to categories mapping file.
        
        Returns
        -------
        :aaindex_category : dict
            Dictionary that maps each AAI record into 1 of 8 categories.
        """
        #if parsed categories file doesn't exist in DATA_DIR then call function
        if not (os.path.isfile(os.path.join(self.pySAR_module_path, DATA_DIR, category_file))):
            self.parse_categories()

        #read categories file and its content
        try:
            with open(os.path.join(self.pySAR_module_path, DATA_DIR, category_file)) as f_out:
                reader = csv.reader(f_out, delimiter="\t")
                d = list(reader)
        except IOError:
            print('Error opening AAIndex1 category file {}.'.format(category_file))

        aaindex_category = {}

        #iterate through all lines in parsed file and store AAI indices as keys
        #   in the output dict and their respective categories as the values of dict
        for i in range(0,len(d)):
          for j in range(0,len(d[i])):
            category_substring = d[i][1].strip()
            category_substring = category_substring.split(" ",1)
            aaindex_category[d[i][0]] = category_substring[0]

        return aaindex_category

    def get_amino_acids(self):
        """
        Get all canonical amino acid letters. The '-' value will also be included
        in the list from this function as it accounts for the abcense of any
        amino acid or gaps in a AAI record.

        Returns
        -------
        :amino_acids : list
            List of all 20 canoniocal amino acid letters as found in each record
            of the AAI database.
        """
        amino_acids = list(self.aaindex_json[list(self.aaindex_json.keys())[0]]["values"].keys())
        amino_acids.sort()  #sort into alphabetical order
        
        return amino_acids

    def get_amino_acids_encoding(self):
        """
        Get one-hot encoding of amino acids.

        Returns
        -------
        :onehot_encoded : np.ndarray
            one hot encoded array of the 20 canonical amino acids.
        """
        all_amino_acids = self.get_amino_acids()
        #convert amino acid letters to np array
        values = np.array(all_amino_acids[1:])    

        #encode amino acids with value between 0 and n_classes-1.
        label_encoder = LabelEncoder()
        integer_encoded = label_encoder.fit_transform(values)

        #encode amino acids as a one-hot numeric array.
        onehot_encoder = OneHotEncoder(sparse=False)
        integer_encoded = integer_encoded.reshape(len(integer_encoded), 1)
        onehot_encoded = onehot_encoder.fit_transform(integer_encoded)

        return onehot_encoded

    def record_codes(self):
       """
       Get list of all AAI index codes/Accession numbers for each record in the database.

       Returns
       -------
       :records : list
         list of record names/index codes for all records in the AAI database.
       """
       records = list(self.aaindex_json.keys())
       records.sort()  #sort into alphabetical order

       return records

    def get_num_records(self):
       """
       Calculate total number of records/indices in the AAI database.

       Returns
       -------
       :len(self.record_codes()) : int
         number of indices/records found in the AAI database.
       """
       return len(self.record_codes())

    def get_record_names(self):
       """
       Return a list of all index descriptions for all records in the AAI database.

       Returns
       -------
       :desc : list
          list of descriptions for all records in the AAI database.
       """
       desc = []

       #iterate through database, appending all descriptions to desc list
       for name in list(self.aaindex_json.values()):
         desc.append(name['description'])

       return desc

    def get_record_from_desc(self, description, all_matches=False):
        """
        Return full AAI database record details from its description. Search
        through the descriptions/names of all records in the database, returning
        record that matches name input parameter. All matches can be returned or
        just the most relevant one, by default just the first match is found, this
        can be changed using the all_matches input param.

        Parameters
        ----------
        :description : str
            AAI database record feature name/description.
        :all_matches : bool (default=false)
            if set to true, all found matches in database returned else if false only the
            first found match is returned.

        Returns
        -------
        :matches : list(dict)
            closest found AAIndex record that matches user input description parameter,
            according to closeness function OR all found relevant AAIndex records.
        """
        matches = []
        all_desc = []

        #iterate through all records in AAI finding matching records with description
        for index, value in self.aaindex_json.items():
            all_desc.append(value['description'].lower())

        #use closeness function to calculate whether description mostly matches record description
        record_matches = get_close_matches(description.lower(), all_desc, cutoff=0.5)

        #if no matches found, return empty list
        if (record_matches==[]):
            return []
        #else iterate through all records in AAI, finding one with most matching description to
        #the input parameter, using the closeness function. Return most found record or all 
        #records, depending on all_matches parameter.
        else:
            for index, value in self.aaindex_json.items():
                if not all_matches:
                    if (record_matches[0] == value['description']):
                        matches.append(self.aaindex_json[index])
                else:
                    for i in range(0,len(record_matches)):
                        if (record_matches[i] == value['description']):
                            matches.append(self.aaindex_json[index])    

        return matches

    def get_values_from_record(self, record_code):
        """
        Return amino acid values of AAI record from its 
        feature/index code/accession number.

        Parameters
        ----------
        :record_code : str
            AAI database record feature code/index code.

        Returns
        -------
        :values : dict
            amino acid values for specified AAI database record.
        """
        #stripping input of whitespace
        try:
            record_code = record_code.strip()
        except:
            raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
                .format(record_code, type(record_code)))

        #check that inputted record_code/Accession num does exist in the AAI database
        if record_code not in (self.record_codes()):
            raise ValueError('Record index {} not found in AAI'.format(record_code))

        #get amino acid values for specified index
        values = (self.aaindex_json[record_code]['values'])

        return values

    def get_ref_from_record(self, record_code):
        """
        Return reference details of a AAI database record from its 
        feaure/index code/accession number.

        Parameters
        ----------
        :record_code : str
            AAI database record feature code/index code.

        Returns
        -------
        :refs : list
            reference details for specifed AAI database record.
        """
        #stripping input of whitespace
        try:
            record_code = record_code.strip()
        except:
            raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
                .format(record_code, type(record_code)))

        #check that inputted record_code does exist in the AAI database
        if record_code not in (self.record_codes()):
                raise ValueError('Record index {} not found in AAI'.format(record_code))

        #get refs from record
        refs = (self.aaindex_json[record_code]['refs'])   

        return refs

    def get_category_from_record(self, record_code):
        """
        Return category of a AAI database record from its 
        feaure/index code/accession number.

        Parameters
        ----------
        :record_code : str
            AAI database record feature code/index code.

        Returns
        -------
        :cat : str
            category of AAI database record according to parsed aaindex_to_category.txt file.
        """
        #stripping input of whitespace
        try:
            record_code = record_code.strip()
        except:
            raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
                .format(record_code, type(record_code)))

        #check that inputted record_code does exist in the AAI database
        if record_code not in (self.record_codes()):
            raise ValueError('Record index {} not in AAI database'.format(record_code))

        #get category from dict
        cat = self.categories[record_code]

        return cat

######################          Getters & Setters          ######################

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def aaindex_filename(self):
        return self._aaindex_filename

    @aaindex_filename.setter
    def aaindex_filename(self, value):
        self._aaindex_filename = value

    @property
    def categories(self):
        return self._categories

    @categories.setter
    def categories(self, value):
        self._categories = value

################################################################################

    def __repr__(self):
        return (json.dumps(self.aaindex_json, indent=4, sort_keys=True))

    def __sizeof__(self):
        """ Return size of AAI database file """
        return os.path.getsize(os.path.isfile(os.path.join(DATA_DIR, self.aaindex_filename)))

    def __getitem__(self, record_code):
        """
        Return full AAI database record details from its feature/index code/Accession number by
        making the whole database subscriptable

        Parameters
        ----------
        :record_code : str
            AAI database record feature index/Accession number.

        Returns
        -------
        :record : dict
            dict of AAI database record.

        Usage
        -----
            aaindex = AAIndex()
            full_record = aaindex['ZIMJ680105']
        """
        #stripping input of whitespace
        try:
            record_code = record_code.strip()
        except:
            raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
                .format(record_code, type(record_code)))

        #check that inputted record_code/Accession num does exist in the AAI database
        if record_code not in (self.record_codes()):
            raise ValueError('Record Index ({}) not found in AAIndex'.format(record_code))
        
        #get full record from parsed JSON
        record = self.aaindex_json[record_code]

        return record

############# Not Finished #############

# class AAIndex2():
#     """ 
#             Python parser for AAindex2: Amino Acid Index Database

#     (Data Format of AAindex2)
#     ************************************************************************
#     *                                                                      *
#     * Each entry has the following format.                                 *
#     *                                                                      *
#     * H Accession number                                                   *
#     * D Data description                                                   *
#     * R PMID                                                               *
#     * A Author(s)                                                          *
#     * T Title of the article                                               *
#     * J Journal reference                                                  *
#     * * Comment or missing                                                 *
#     * M rows = ARNDCQEGHILKMFPSTWYV, cols = ARNDCQEGHILKMFPSTWYV           *
#     *   AA                                                                 *
#     *   AR RR                                                              *
#     *   AN RN NN                                                           *
#     *   AD RD ND DD                                                        *
#     *   AC RC NC DC CC                                                     *
#     *   AQ RQ NQ DQ CQ QQ                                                  *
#     *   AE RE NE DE CE QE EE                                               *
#     *   AG RG NG DG CG QG EG GG                                            *
#     *   AH RH NH DH CH QH EH GH HH                                         *
#     *   AI RI NI DI CI QI EI GI HI II                                      *
#     *   AL RL NL DL CL QL EL GL HL IL LL                                   *
#     *   AK RK NK DK CK QK EK GK HK IK LK KK                                *
#     *   AM RM NM DM CM QM EM GM HM IM LM KM MM                             *
#     *   AF RF NF DF CF QF EF GF HF IF LF KF MF FF                          *
#     *   AP RP NP DP CP QP EP GP HP IP LP KP MP FP PP                       *
#     *   AS RS NS DS CS QS ES GS HS IS LS KS MS FS PS SS                    *
#     *   AT RT NT DT CT QT ET GT HT IT LT KT MT FT PT ST TT                 *
#     *   AW RW NW DW CW QW EW GW HW IW LW KW MW FW PW SW TW WW              *
#     *   AY RY NY DY CY QY EY GY HY IY LY KY MY FY PY SY TY WY YY           *
#     *   AV RV NV DV CV QV EV GV HV IV LV KV MV FV PV SV TV WV YV VV        *
#     * //                                                                   *
#     ************************************************************************

#     Attributes
#     ----------
#     :aaindex_filename : str (default = 'aaindex2')
#         local filename of aaindex2 file that class will import the database
#         from in the data directory, default value of 'aaindex2' is reccomended.
#     :download_using : str (default = "ftp")
#         decide to download AAI database using ftp or https, ftp used by default.

#     Methods
#     -------
#     parse_aaindex():
#         parse AAI database into JSON format.

#     References
#     ----------
#     [1]: Nakai, K., Kidera, A., and Kanehisa, M.;  Cluster analysis of
#         amino acid indices for prediction of protein structure and
#         function.  Protein Eng. 2, 93-100 (1988)
#     [2]: Tomii, K. and Kanehisa, M.;  Analysis of amino acid indices and
#         mutation matrices for sequence comparison and structure
#         prediction of proteins.  Protein Eng. 9, 27-36 (1996).
#     [3]: Kawashima, S., Ogata, H., and Kanehisa, M.;  AAindex: amino acid
#         index database.  Nucleic Acids Res. 27, 368-369 (1999).
#     [4]: Kawashima, S. and Kanehisa, M.;  AAindex: amino acid index
#         database.  Nucleic Acids Res. 28, 374 (2000).
#     """

#     def __init__(self, aaindex_filename='aaindex2', download_using='ftp'):  

#         self.aaindex_filename = aaindex_filename
#         self.download_using = download_using

#         #download AAI database using ftp or https
#         if self.download_using=='ftp':
#             self.url = "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex2"
#         elif self.download_using=='http' or self.download_using=='https':
#             self.url = "https://www.genome.jp/ftp/db/community/aaindex/aaindex2"
#         else:
#             self.url = "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex2"

#         #if AAIndex database not found in data directory then call download method
#         if not (os.path.isfile(os.path.join(self.pySAR_module_path,DATA_DIR,self.aaindex_filename))):
#             self.download_aaindex()

#         #if parsed json of AAIndex already in file then read it and return  <- Not working at the moment
#         # if (os.path.isfile(os.path.join(self.pySAR_module_path,'data', self.aaindex_filename +'.json'))):
#         #     with open(os.path.join(self.pySAR_module_path,DATA_DIR, self.aaindex_filename +'.json')) as aai_json:
#         #         self.aaindex_json = json.load(aai_json)
#         # else:
#         #     #parse AAIndex database file into JSON format
#         #     self.aaindex_json = self.parse_aaindex()

#         # self.categories = self.get_all_categories()

#     def parse_aaindex(self):
#         """
#         Parse AAI database into JSON format. Each AAI record will be indexed by
#         its feature code/index code, and will be in the format as shown in the
#         example above. The file will be stored in a json file called according
#         to (self.aaindex_filename+'.json') variable.

#         Returns
#         -------
#         : aaindex_json : dict
#           parsed AAI database in dict form.
#         """
#         #initialise keys of AAI database
#         template_dict = {"H":[],
#                          "D":[],
#                          "R":[],
#                          "A":[],
#                          "*":[],
#                          "T":[],
#                          "J":[],
#                          "M":[]}

#         #open AAI file for reading and parsing, by default it should be stored in DATA_DIR
#         try:
#             # tmp_filepath = os.path.join(DATA_DIR,self.aaindex_filename)
#             tmp_filepath = os.path.join(self.pySAR_module_path, 'data',self.aaindex_filename)
#             f = open(tmp_filepath,'r')
#         except IOError:
#             print('Error opening file, check filename = {} and is stored in \
#                     {} directory.'.format(self.aaindex_filename, DATA_DIR))

#         #read lines of file
#         lines = f.readlines()
#         f.close()

#         clean_up_pattern = re.compile("\"")

#         #initilaise parsed AAI database dictionary
#         aaindex_json = {}
#         current_dict = copy.deepcopy(template_dict)

#         '''
#         iterate through each line in the AAI database, parsing each record and its
#         amino acid values into its own entry in the dictionary. Each index/record
#         is seperated by a '//'. Remove any duplicate records, set any missing
#         ('-') or NA amino acid values to 0. Store resulting dict into aaindex_json
#         instance variable.
#         '''
#         for l in lines:

#             if l.startswith("//"):

#                 # deal with meta data
#                 name = " ".join(current_dict["H"])
#                 name = clean_up_pattern.sub("'",name)

#                 description = " ".join(current_dict["D"])
#                 description = clean_up_pattern.sub("'",description)

#                 citation = "{} '{}' {}".format(" ".join(current_dict["A"]),
#                                                " ".join(current_dict["T"]),
#                                                " ".join(current_dict["J"]))

#                 citation = citation + "; Kawashima, S. and Kanehisa, M. \
#                     'AAindex: amino acid index database.'  Nucleic Acids Res. 28, 374 (2000)."

#                 citation = clean_up_pattern.sub("'",citation)

#                 notes = " ".join(current_dict["*"])
#                 notes = clean_up_pattern.sub("'",notes)

#                 # parse amino acid data
#                 aa_lines = current_dict["I"]

#                 aa_names = aa_lines[0].split()
#                 row_0_names = [aa.split("/")[0] for aa in aa_names]
#                 row_1_names = [aa.split("/")[1] for aa in aa_names]

#                 row_0_values = aa_lines[1].split()
#                 row_1_values = aa_lines[2].split()

#                 values = {}
#                 for i in range(len(row_0_values)):
#                     try:
#                         values[row_0_names[i]] = float(row_0_values[i])
#                     except ValueError:
#                         values[row_0_names[i]] = "NA"

#                     try:
#                         values[row_1_names[i]] = float(row_1_values[i])
#                     except ValueError:
#                         values[row_1_names[i]] = "NA"

#                 # look for duplicate name entries
#                 try:
#                     aaindex_json[name]
#                     err = "duplicate value name ({})".format(name)
#                     raise ValueError('Duplicate AAI Record name found')
#                 except KeyError:
#                     pass

#                 aaindex_json[name] = {"description":description,
#                                   "refs":citation,
#                                   "notes":notes,
#                                   "values":values}

#                 current_dict = copy.deepcopy(template_dict)
#                 continue

#             this_entry = l[0]
#             if l[0] != " ":
#                 current_entry = this_entry

#             current_dict[current_entry].append(l[1:].strip())

#         categories = self.get_all_categories()

#         #append '-' to each aa index entry to account for missing AA in a protein sequence
#         for index in aaindex_json:

#             #add category to aaindex file
#             aaindex_json[index]['category'] = categories[index]
#             aaindex_json[index]['values']['-'] = 0

#             #set any NA amino acid values to 0
#             for val in aaindex_json[index]['values']:
#                 if aaindex_json[index]['values'][val] == 'NA':
#                     aaindex_json[index]['values'][val] = 0

#         #save parsed dictionary into JSON format to DATA_DIR
#         with open((os.path.join(self.pySAR_module_path, 'data', self.aaindex_filename + '.json')),'w') as output_F:
#           json.dump(aaindex_json, output_F, indent=4, sort_keys=True)

#         return aaindex_json

#     def download_aaindex(self, save_dir=DATA_DIR):
#         """
#         If AAI database not found in DATA directory, then it will be downloaded from
#         the dedicated FTP or HTTPS server from https://www.genome.jp/aaindex/.
#         FTP is the default method used for downloading the database with the URL:
#         "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1". If you want to
#         download by https, set the 'download_using' instance variable to 'https'.

#         Parameters
#         ----------
#         : save_dir : str (default = DATA_DIR)
#             Directory to save the AAI database to. By default it wil be stored in
#             the global var DATA_DIR = 'data'.
#         """
#         #fetch AAI database from URL if not present in DATA_DIR
#         try:
#             if not(os.path.isfile(os.path.join(self.pySAR_module_path, DATA_DIR, self.aaindex_filename))):
#                 try:
#                     with closing(request.urlopen(self.url)) as r:
#                         with open((os.path.join(self.pySAR_module_path, DATA_DIR,self.aaindex_filename)), 'wb') as f:
#                             shutil.copyfileobj(r, f)
#                     print('AAIndex2 successfully downloaded.')
#                 except requests.exceptions.RequestException:
#                     print('Error downloading or exporting AAIndex from the url {}.'.format(self.url))
#             else:
#                 pass    #AAIndex already in folder
#         except:
#             raise OSError('Save Directory does not exist: {}.'.format(save_dir))

# class AAIndex3():
#     pass



################################################################################
#################              PyDSP Module Tests             #################
################################################################################

import os
import pandas as pd
import numpy as np
import pySAR.aaindex as aaindex
import pySAR.pyDSP as pyDSP_
import pySAR.pySAR as pysar
import unittest

@unittest.skip("")
class pyDSPTests(unittest.TestCase):

    def setUp(self):
        """  Import the 4 config files for each of the 4 datasets used for testing the pyDSP methods. """        
        #array of config files for each test dataset
        config_path = os.path.join('tests','test_config')
        self.all_config_files = [
            os.path.join(config_path, "test_thermostability.json"), 
            os.path.join(config_path, "test_enantioselectivity.json"),
            os.path.join(config_path, "test_absorption.json"), 
            os.path.join(config_path, "test_localization.json")
        ]

    def test_pyDSP(self):
        """ Test class input parameters and attributes. """
#1.)   
        # encoded_seq1 = self.pySAR.get_aai_enoding(test_aaindices1)            
        pyDSP = pyDSP_.PyDSP(dsp_config=self.all_config_files[0]) #test_thermostability
        self.assertEqual(pyDSP.spectrum, "power")
        self.assertEqual(pyDSP.window_type, "hamming")
        self.assertIsInstance(pyDSP.window, np.ndarray)
        self.assertIsNone(pyDSP.filter_)
        self.assertEqual(pyDSP.encoded_sequences.shape, (self.pySAR.num_seqs, self.pySAR.seq_len))
        self.assertEqual(pyDSP.num_seqs, self.pySAR.num_seqs)
        self.assertEqual(pyDSP.signal_len, self.pySAR.seq_len)
        # self.assertEqual(pyDSP.fft_power.shape, encoded_seq1.shape)
        self.assertTrue(pyDSP.fft_power.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_real.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_real.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_abs.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_abs.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_imag.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_imag.dtype, 'float64')
        self.assertTrue(pyDSP.spectrum_encoding.any() == pyDSP.fft_power.any())
        # self.assertEqual(pyDSP.fft_freqs.shape,encoded_seq1.shape)
#2.)
        pyDSP = pyDSP_.PyDSP(dsp_config=self.all_config_files[1]) #test_enantioselectivity
        self.assertEqual(pyDSP.spectrum, "power")
        self.assertEqual(pyDSP.window_type, "hamming")
        self.assertIsInstance(pyDSP.window, np.ndarray)
        self.assertIsNone(pyDSP.filter_)
        self.assertEqual(pyDSP.encoded_sequences.shape, (self.pySAR.num_seqs,self.pySAR.seq_len))
        self.assertEqual(pyDSP.num_seqs, self.pySAR.num_seqs)
        self.assertEqual(pyDSP.signal_len, self.pySAR.seq_len)
        # self.assertEqual(pyDSP.fft_power.shape, encoded_seq1.shape)
        self.assertTrue(pyDSP.fft_power.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_real.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_real.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_abs.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_abs.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_imag.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_imag.dtype, 'float64')
        self.assertTrue(pyDSP.spectrum_encoding.any() == pyDSP.fft_power.any())
        # self.assertEqual(pyDSP.fft_freqs.shape,encoded_seq1.shape)

#3.) 
        pyDSP = pyDSP_.PyDSP(dsp_config=self.all_config_files[2], protein_seqs="") #test_absorption
        #pre-encode sequences
        self.assertEqual(pyDSP.spectrum, "power")
        self.assertEqual(pyDSP.window_type, "hamming")
        self.assertIsInstance(pyDSP.window, np.ndarray)
        self.assertIsNone(pyDSP.filter_)
        self.assertEqual(pyDSP.encoded_sequences.shape, (self.pySAR.num_seqs, self.pySAR.seq_len))
        self.assertEqual(pyDSP.num_seqs, self.pySAR.num_seqs)
        self.assertEqual(pyDSP.signal_len, self.pySAR.seq_len)
        # self.assertEqual(pyDSP.fft_power.shape, encoded_seq1.shape)
        self.assertTrue(pyDSP.fft_power.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_real.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_real.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_abs.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_abs.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_imag.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_imag.dtype, 'float64')
        self.assertTrue(pyDSP.spectrum_encoding.any() == pyDSP.fft_power.any())
        # self.assertEqual(pyDSP.fft_freqs.shape,encoded_seq1.shape)
#4.) 
        pyDSP = pyDSP_.PyDSP(dsp_config=self.all_config_files[3]) #test_localization
        self.assertEqual(pyDSP.spectrum, "power")
        self.assertEqual(pyDSP.window_type, "hamming")
        self.assertIsInstance(pyDSP.window, np.ndarray)
        self.assertIsNone(pyDSP.filter_)
        self.assertEqual(pyDSP.encoded_sequences.shape, (self.pySAR.num_seqs, self.pySAR.seq_len))
        self.assertEqual(pyDSP.num_seqs, self.pySAR.num_seqs)
        self.assertEqual(pyDSP.signal_len, self.pySAR.seq_len)
        # self.assertEqual(pyDSP.fft_power.shape, encoded_seq1.shape)
        self.assertTrue(pyDSP.fft_power.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_real.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_real.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_abs.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_abs.dtype, 'float64')
        # self.assertEqual(pyDSP.fft_imag.shape, encoded_seq1.shape)
        self.assertEqual(pyDSP.fft_imag.dtype, 'float64')
        self.assertTrue(pyDSP.spectrum_encoding.any() == pyDSP.fft_power.any())
        # self.assertEqual(pyDSP.fft_freqs.shape,encoded_seq1.shape)
#3.)
        with self.assertRaises(OSError, msg='OS Error raised, invalid config file path given.'):
            pyDSP = pyDSP_.PyDSP(dsp_config="blahblahblah")

        with self.assertRaises(TypeError, msg='Type Error raised, invalid config file path data type given.'):
            pyDSP = pyDSP_.PyDSP(dsp_config=4.21)

    def test_preprocessing(self):
        """ Testing preprocessing functionality of pyDSP class. """
        test_aaindices1 = "COHE430101"
#1.)
        encoded_seq1 = self.pySAR.get_aai_enoding(test_aaindices1)
        pyDSP = pyDSP_.PyDSP(encoded_seq1)
        pyDSP.pre_processing()
        self.assertTrue(np.all((pyDSP.fft_power==0)))
        self.assertTrue(np.all((pyDSP.fft_real==0)))
        self.assertTrue(np.all((pyDSP.fft_imag==0)))
        self.assertTrue(np.all((pyDSP.fft_abs==0)))
        self.assertFalse(np.isnan(pyDSP.encoded_sequences).any(), 'Sequences should not contain null values.')
#2.)

    def test_protein_spectra(self):
        """ Testing getting protein spectra from encoded protein sequences. """
        test_aaindices1 = "COHE430101"
#1.)
        encoded_seq1 = self.pySAR.get_aai_enoding(test_aaindices1)
        pyDSP = pyDSP_.PyDSP(encoded_seq1)
        self.assertTrue(pyDSP.fft_power.dtype, "complex128")
        self.assertTrue(pyDSP.fft_real.dtype, "complex128")
        self.assertTrue(pyDSP.fft_imag.dtype, "complex128")
        self.assertTrue(pyDSP.fft_abs.dtype, "complex128")

    def test_max_freq(self):
        """ Testing max frequency functionality. """
#1.)
        test_aaindices1 = "COHE430101"
        encoded_seq1 = self.pySAR.get_aai_enoding(test_aaindices1)
        pyDSP = pyDSP_.PyDSP(encoded_seq1)




################################################################################
#################                     pySAR                    #################
################################################################################

import pandas as pd
import numpy as np
import os
import sys
from difflib import get_close_matches
import json
from json import JSONDecodeError

from .globals_ import OUTPUT_DIR, OUTPUT_FOLDER, DATA_DIR
from .aaindex import AAIndex
from .model import Model
from .pyDSP import PyDSP
from .evaluate import Evaluate
from .utils import *
from .plots import plot_reg
from .descriptors_ import Descriptors

class PySAR():
    """
    The PySAR class is the main class for the pySAR software. The class allows for
    the encoding of protein sequences via a plethora of techniques, mainly via
    AAI Indices and or strucutrual and physiochemical protein descriptors that
    are then used as features in the building of a predictive model created to
    map the protein sequences to a sought-after activity value (activity attribute),
    this is known as a sequence activity relationship (SAR). Three main encoding
    strategies are possible in the class, namely using AAI Indices or protein
    descriptors as well as AAI Indices + Descriptors. Additionally, the protein sequences
    can be encoded using Digital Signal Processing (DSP) techniques, mainly through the use
    of informational protein spectra, this is achieved via the PyDSP class in the software.
    The class accepts strings or lists of AAI Indices or descriptors and then passes 
    these through a pipeline to get the required encoding of the respective sequences. 
    The calculated encodings of the sequences are used as features in the building 
    of the predictive model that will predict the acitivty values for new unseen 
    protein sequences. After the encoding process, various metrics will be captured 
    and stored in a local output folder according to the OUTPUT_FOLDER global var 
    as well as a regression plot showing how well the model, and the selected protein feature 
    attributes, fit to the test data of unseen protein sequences.

    Attributes
    ----------    
    :config_file (str): 
        path to configuration file.
        
    Methods
    -------
    read_data():
        read dataset of protein sequences.
    preprocessing():
        pre-process / clean protein sequence dataset.
    get_aai_encoding(indices):
        get AAI encoding for user inputted index/indices.
    encode_aai(indices=None):
        get encoded protein sequences according to user inputted index/indices, applying DSP if applicable. 
    get_descriptor_encoding(descriptors):
        calculate user inputted descriptor/descriptors according to user input.
    encode_desc(descriptor=None):
        get encoded protein sequences according to user inputted descriptor/descriptors.
    encode_aai_desc(indices=None, descriptors=None)
        get encoded protein sequences according to user inputted AA index/indices and descriptor/descriptors.
    output_results(results):
        print out the predictive model parameters/attributes and its results.
    get_seqs():
        return protein sequences
    get_activity():
        return array of protein sequences' activity values.
    """
    def __init__(self, config_file=""):

        self.config_file = config_file
        self.params = {}

        config_filepath = ""
        #open json config file       
        if not isinstance(config_file, str) or config_file is None:
            raise TypeError('JSON config file must be a filepath of type string, not of type {}'.format(type(config_file)))
        if os.path.isfile(self.config_file):
            config_filepath = self.config_file
        elif os.path.isfile(os.path.join('config', self.config_file)):
            config_filepath = os.path.join('config', self.config_file)
        else:
            raise OSError('JSON config file not found at path: {}.'.format(config_filepath))
        try:
            with open(config_filepath) as f:
                self.params = json.load(f)
        except JSONDecodeError as e:
            print('Error parsinf config JSON file: {}.'.format(config_filepath))
            sys.exit()

        #dataset parameters
        self.dataset = self.params["dataset"][0]["dataset"]
        self.sequence_col = self.params["dataset"][0]["sequence_col"]
        self.activity = self.params["dataset"][0]["activity"]

        #model parameters
        self.algorithm = self.params["model"][0]["algorithm"]
        self.parameters = self.params["model"][0]["parameters"]
        self.test_split = self.params["model"][0]["test_split"]

        #aai parameters
        self.aai_indices = None

        #descriptors parameters
        # self.descriptors = self.params["descriptors"]
        self.descriptors = None

        #pyDSP parameters
        self.use_dsp = self.params["pyDSP"][0]["use_dsp"]
        if (self.use_dsp):
            self.spectrum = self.params["pyDSP"][0]["spectrum"]
            self.window = self.params["pyDSP"][0]["window"]
            self.filter = self.params["pyDSP"][0]["filter"]
            self.convolution = self.params["pyDSP"][0]["convolution"]

        #import and read dataset
        self.data = self.read_data()

        #array of protein sequences
        self.sequences = self.data[self.sequence_col]

        #pre-process dataset and protein sequences
        self.preprocessing()

        #get number of rows and cols of dataset
        self.num_seqs = len(self.data[self.sequence_col])
        self.seq_len = len(max(self.data[self.sequence_col], key=len))

        #create instance of AAIndex class
        self.aaindex = AAIndex()

        #create instance of Descriptors class
        self.descriptor = Descriptors(self.config_file, protein_seqs=self.data[self.sequence_col])

        #create instance of model class of type specified by algorithm
        self.model = Model(self.algorithm, parameters=self.parameters, test_split=self.test_split)

        #updating algorithm attribute
        self.algorithm = repr(self.model)

        #create output folder to store all model assets and results
        create_output_dir()

    def read_data(self):
        """
        Read in dataset according to name from 'dataset' attribute. By default
        the dataset should be stored in DATA_DIR.
        
        Parameters
        ----------
        :self (PySAR object): 
            instance of PySAR class.

        Returns
        -------
        :data (pd.DataFrame): 
            dataframe of dataset      
        """
        filepath = ""
        #read in dataset csv if found in path, if not raise error
        if (os.path.isfile(os.path.join(DATA_DIR,self.dataset))):
            filepath = os.path.join(DATA_DIR,self.dataset)
        elif os.path.isfile(self.dataset):
            filepath = self.dataset
        else:
            raise OSError('Dataset filepath is not correct: {}'.format(filepath))

        #read in dataset csv
        try:
            data = pd.read_csv(filepath, sep=",", header=0)
            return data
        except:
            raise IOError('Error opening dataset file: {}'.format(filepath))

    def preprocessing(self):
        """
        Pre-process protein sequences in dataset. Validate column names, check
        for invalid amino acids in sequences, remove any gaps in sequence and 
        remove any NAN or +/- infinity values.

        Parameters
        ----------
        :self (PySAR object): 
            instance of PySAR class.

        Returns
        -------
        None
        """
        #get closest match for sequence column name in dataset
        sequence_col_matches = get_close_matches(self.sequence_col, self.data.columns, cutoff=0.4)

        #set sequence col to the first match found, else raise error
        if sequence_col_matches!=[]:
            self.sequence_col = sequence_col_matches[0]
        else:
            raise ValueError('Sequence Column ({}) not present in dataset columns - {} /n '.
                format(self.sequence_col, self.data.columns))

        #remove any gaps found in sequences in dataset
        self.data[self.sequence_col] = remove_gaps(self.data[self.sequence_col])

        #verify no invalid amino acids found in sequences, if so then raise error
        invalid_seqs = valid_sequence(self.get_seqs())
        if invalid_seqs!=None:
            raise ValueError('Invalid Amino Acids found in protein sequence dataset: {}'.
                format(invalid_seqs))

        #get closest match for activity column name in dataset
        activity_matches = get_close_matches(self.activity, self.data.columns, cutoff=0.4)

        #set activity col to the first match found, else raise error
        if activity_matches!=[]:
            self.activity = activity_matches[0]
        else:
            raise ValueError('Activity Column ({}) not present in dataset columns  - {} /n '.
                format(self.activity,list(self.data.columns)))

        #remove any +/- infinity values or any Null/NAN's from activity values
        self.data[self.activity].replace([np.inf,-np.inf], np.nan)
        self.data[self.activity].fillna(0,inplace=True)

    def get_aai_encoding(self, indices=None):
        """ 
        Get AAI index encoding values for index specified by indices for each amino
        acid in each of the protein sequences in dataset. If multiple indices
        specified then encode protein sequences with each index and concatenate.

        Parameters
        ----------
        :indices : str/list (default = None):
            string or list of AAI indices.
        
        Returns
        -------
        :encoded_seqs : np.ndarray:
            array of the encoded protein sequences in dataset via user input index/indices.
        """
        #validate AAI indices are present in the input parameter, if not raise error
        if indices == None or indices == "":
            raise ValueError('AAI indices input parameter cannot be None or empty.')

        #check input indices is of correct type (str/list), if not raise type error
        if (not isinstance(indices, str) and (not isinstance(indices, list))):
            raise TypeError("Input indices parameter must be a string or list, got {}".format(type(indices)))

        encoded_indices = []

        #if single index input passed into parameter
        if not isinstance(indices, list):

            #get dict of amino acid encoding values for index specified by indices
            encoded_aai = self.aaindex[indices]['values']

            #initialise temp arrays to store encoded sequences
            temp_seq_vals = []
            temp_all_seqs = []

            #iterate through each protein sequence and amino acid, getting the AAI index encoding value
            for protein in range(0, len(self.data[self.sequence_col])):
                for aa in self.data[self.sequence_col][protein]:
                    temp_seq_vals.append(encoded_aai[aa])

                #append encoding and reset temp array
                temp_all_seqs.append(temp_seq_vals)
                temp_seq_vals = []

            #zero-pad encoding list so that sequences are all the same length
            temp_all_seqs = zero_padding(temp_all_seqs)

            #convert list of lists into array
            temp_all_seqs = np.array(temp_all_seqs, dtype="float32")

            encoded_seqs = temp_all_seqs

            return encoded_seqs

        #if multiple indices (list) passed into input parameter
        else:

            #create zeros numpy array to store encoded sequence output
            encoded_aai_ = np.zeros((self.num_seqs, self.seq_len))

            #if multiple indices used then calcualte AAI index encoding for each and
            #then concatenate after each calculation
            for ind in range(0,len(indices)):
                encoded_aai = self.aaindex[indices[ind]]['values']

                #initialise temp arrays to store encoded sequences
                temp_seq_vals = []
                temp_all_seqs = []

                #iterate through each protein sequence and amino acid, getting the AAI index encoding value
                for protein in range(0, len(self.data[self.sequence_col])):
                    for aa in self.data[self.sequence_col][protein]:
                        temp_seq_vals.append(encoded_aai[aa])

                    #append encoding and reset temp array
                    temp_all_seqs.append(temp_seq_vals)
                    temp_seq_vals = []

                #zero-pad encoding list so that sequences are all the same length
                temp_all_seqs = zero_padding(temp_all_seqs)

                #convert list of lists into array
                temp_all_seqs = np.array(temp_all_seqs, dtype="float32")

                #in first iteration through indices set encoded_aai_ to zero-initialised 
                # np array, else concatenate to array in previous iteration
                if ind == 0:
                    encoded_aai_ = temp_all_seqs
                else:
                    encoded_aai_ = np.concatenate((encoded_aai_,temp_all_seqs), axis=1)

            encoded_seqs = encoded_aai_

            return encoded_seqs

    def encode_aai(self, indices=None, show_plot=False):
        """
        Encode using AAI indices. If multiple indices input then calculate each
        and concatenate them. Build predictive model from AAI feature data. The
        resulting model assets and its results will be exported to the directory
        pointed to by the global variable OUTPUT_DIR. If use_dsp config parameter 
        is true then pass AAI Indices through a DSP transformation specified by 
        the config's DSP parameters (spectrum, window, filter & convolution) via the 
        PyDSP module.

        Parameters
        ----------
        :indices : str/list (default = None)
            string or list of indices from the AAI.
        :show_plot : bool (default = False)
            display regression plot of best predictive model. If False then the plot
            will just be saved to the output folder, else it'll be displayed & saved.

        Returns
        -------
        :aai_df : pd.Series
            pandas series storing metrics and results of encoding.
        """
        #validate AAI indices are present in the input parameter
        if indices == None or indices == "" or indices == []:
            raise ValueError('AAI indices input parameter cannot be None or empty.')

        #check input indices is of correct type (str/list), if not raise type error
        if ((not isinstance(indices, str)) and (not isinstance(indices, list))):
            raise TypeError("Input indices parameter must be a string or list, got {}".format(type(indices)))

        self.aai_indices = indices

        #pandas series to store all output results
        aai_df = pd.Series(index=['Index', 'Category','R2', 'RMSE', 'MSE', 'RPD',
            'MAE', 'Explained Variance'], dtype='object')

        #get AAI index encodings specified by indices input parameter
        encoded_seqs = self.get_aai_encoding(indices)

        #if use_dsp true then get protein spectra from AAI indices using PyDSP class,
        #else use the AAI indices encoding's themselves as the feature data (X)
        if self.use_dsp:
            #if input spectrum is none or empty, raise error.
            if self.spectrum == None or self.spectrum == "":
                raise ValueError('Spectrum cannot be None or empty, got {}'.format(self.spectrum))
            pyDSP = PyDSP(self.config_file, protein_seqs=encoded_seqs)
            pyDSP.encode_seqs()
            X = pd.DataFrame(pyDSP.spectrum_encoding)
        else:
            X = pd.DataFrame(encoded_seqs)  #no DSP applied to encoded sequences

        #get training and test dataset split
        X_train, X_test, Y_train, Y_test = self.model.train_test_split(X,
            self.get_activity(), test_size=self.test_split)

        #fit predictive model
        self.model.fit()

        #predict activity values for test data
        Y_pred = self.model.predict()

        #create instance of Evaluate class which will get all the evaluation metrics
        eval = Evaluate(Y_test, Y_pred)

        #get categories for all indices in self.aai_indices
        if isinstance(self.aai_indices, list):
            index_cat = ""
            for i in range(0,len(self.aai_indices)):
                index_cat += index_cat + ', '+ \
                    self.aaindex.get_category_from_record(self.aai_indices[i])
        else:
            index_cat = self.aaindex.get_category_from_record(self.aai_indices)

        #set results Series variables
        aai_df['Index'] = indices
        aai_df['Category'] = str(index_cat)
        aai_df['R2'] = eval.r2
        aai_df['RMSE'] = eval.rmse
        aai_df['MSE'] = eval.mse
        aai_df['RPD'] = eval.rpd
        aai_df['MAE'] = eval.mae
        aai_df['Explained Variance'] = eval.explained_var

        #print out results from encoding
        self.output_results(aai_df)

        #plot regression plot for predictive model
        plot_reg(Y_test, Y_pred, eval.r2, show_plot)

        #save results of encoding to output folder specified by OUTPUT_FOLDER
        save_results(aai_df, 'aai_encoding')

        #reset aai_indices instance variable
        # self.aai_indices = ""

        return aai_df

    def get_descriptor_encoding(self, descriptors=None):
        """
        Calculate inputted descriptor(s) requried for the encoding process.
        Get closest match to user inputted descriptor using difflib library.
        If a single descriptor is input then calculate it and return, if list of
        descriptors input then calculate descriptor values for each and concatenate.

        Parameters
        ----------
        :descriptors : str/list (default=None)
            string or list of protein descriptors

        Returns
        -------
        :encoded_desc : pd.DataFrame
            pandas dataFrame of calculated descriptor values according to user
            inputted descriptor(s).
        """
        #raise error if no descriptors specified in input
        if descriptors == None or descriptors == "" or descriptors == []:
            raise ValueError('Descriptors input parameter cannot be None or empty.')

        #check input descriptor is of correct type (str), if not raise type error
        if (not isinstance(descriptors, str) and (not isinstance(descriptors, list))):
            raise TypeError("Input descriptor parameter must be a string or list, got {}".format(type(descriptors)))

        #create instance of Descriptors class using data in instance variable
        descr = Descriptors(self.config_file, protein_seqs=self.data[self.sequence_col])

        #get closest valid available descriptors from input descriptor parameter,
        #   if a list of descriptors passed in as the input parameter then get
        #       all valid descriptors in list
        if isinstance(descriptors, list):
            for de in range(0,len(descriptors)):
                desc_matches = get_close_matches(descriptors[de],
                    descr.valid_descriptors(), cutoff=0.4)
                descriptors[de] = desc_matches[0]
                if descriptors[de] == []:
                    raise ValueError('No approximate descriptor found from one entered: {}'.format(de))
        else:
            desc_matches = get_close_matches(descriptors,descr.valid_descriptors(), cutoff=0.4)
            descriptors = desc_matches[0]
            if descriptors == []:
                raise ValueError('No approximate descriptor found from one entered: {}'.format(descriptors))

        #initialise temp lists and DF to store encoded descriptor values
        encoded_desc_temp = []
        encoded_desc_vals = []
        encoded_desc_temp = pd.DataFrame()

        #if single descriptor passed in, not a list, get descriptor values for sequences
        if not isinstance(descriptors, list):
            encoded_desc_temp = descr.get_descriptor_encoding(descriptors)

            #raise value error if descriptor is empty
            if (encoded_desc_temp.empty):
                raise ValueError('Descriptors {} cannot be empty or None'.format(descriptors))

            encoded_desc = encoded_desc_temp

            return encoded_desc

        #if list of descriptors passed as input, iterate and get each descriptors' values
        else:
            for d in range(0,len(descriptors)):
              encoded_desc_temp = descr.get_descriptor_encoding(descriptors[d])

              #raise value error if descriptor is empty
              if (encoded_desc_temp.empty):
                  raise ValueError('Descriptors {} cannot be empty or None'.format(descriptors[d]))
                    # self.descriptors[d]))

              encoded_desc_vals.append(encoded_desc_temp)
              encoded_desc_temp = pd.DataFrame()   #reset to empty dataframe

            #concatenate dataframes of descriptors
            encoded_desc = pd.concat(encoded_desc_vals, axis=1)

        return encoded_desc

    def encode_desc(self, descriptor=None, show_plot=False):
        """
        Encode protein sequences using protein physiochemical and or structural
        descriptors and build predictive model from the descriptor feature data.
        If multiple descriptors input then calculate each and concatenate them.
        The resulting model assets and its results will be exported to the
        directory pointed to by the global variable OUTPUT_DIR.

        Parameters
        ----------
        :descriptor : str/list (default = None)
            string or list of protein descriptors
        :show_plot : bool (default = False)
            display regression plot of best predictive model. If False then the plot
            will just be saved to the output folder, else it'll be displayed & saved.

        Returns
        -------
        :desc_df : pd.Series
            pandas series storing metrics and results of encoding.
        """
        #raise error if no descriptor specified in input
        if descriptor == None or descriptor == "":
            raise ValueError('No descriptors have been input, descriptors attribute is empty.')

        #check input descriptor is of correct type (str), if not raise type error
        if not isinstance(descriptor, str):
            raise TypeError("Input descriptor parameter must be a string, got {}".format(type(descriptor)))

        self.descriptors = descriptor

        #create instance of Descriptors class using data in instance variable
        descr = Descriptors(self.config_file, protein_seqs=self.data[self.sequence_col])

        #get closest matching descriptor from descriptor input parameter
        desc_matches = get_close_matches(self.descriptors, descr.valid_descriptors(), cutoff=0.4)
        if desc_matches!=[]:
            self.descriptors = desc_matches[0]  #set desc to closest descriptor match found
        else:
            raise ValueError('Could not find a match for the input descriptor in available valid descriptors:\n {}.'.
                format(self.descriptors, descr.valid_descriptors()))

        #pandas series to store all output results
        desc_df = pd.Series(index=['Descriptor','Group', 'R2', 'RMSE', 'MSE',
            'RPD', 'MAE', 'Explained Variance'], dtype='object')

        # if isinstance(descriptors, list):
        #     for de in range(0,len(descriptors)):
        #         desc_matches = get_close_matches(descriptors[de],
        #             descr.valid_descriptors(),cutoff=0.4)
        #         descriptors[de] = desc_matches[0]
        #         if descriptors[de] == []:
        #             raise ValueError('No approximate descriptor found from one entered: {}'.format(de))
        # else:
        #     desc_matches = get_close_matches(descriptors,descr.valid_descriptors(),cutoff=0.4)
        #     descriptors = desc_matches[0]
        #     if descriptors == []:
        #         raise ValueError('No approximate descriptor found from one entered: {}'.format(descriptors))

        #set training data (X) to descriptor-encoded protein sequences
        X = self.get_descriptor_encoding(descriptors=self.descriptors)

        #get training and test dataset split
        X_train, X_test, Y_train, Y_test = self.model.train_test_split(X,
            self.get_activity(), test_size=self.test_split)

        #fit predictive model
        self.model.fit()

        #predict activity values for test data
        Y_pred = self.model.predict()

        #create instance of Evaluate class which will get all the evaluation metrics
        eval = Evaluate(Y_test, Y_pred)

        desc_group = ""

        #get groups for all descriptors in self.desciptors
        if isinstance(self.descriptors, list):
            for i in range(0,len(self.descriptors)):
                desc_group += desc_group + ', ' + \
                    descr.descriptor_groups[self.descriptors[i]]
        else:
            desc_cat = descr.descriptor_groups[self.descriptors]

        #set results Series variables
        desc_df['Descriptor'] = descriptor
        desc_df['Group'] = desc_cat
        desc_df['R2'] = eval.r2
        desc_df['RMSE'] = eval.rmse
        desc_df['MSE'] = eval.mse
        desc_df['RPD'] = eval.rpd
        desc_df['MAE'] = eval.mae
        desc_df['Explained Variance'] = eval.explained_var

        #ensure aai indices attribute doesn't show up in output results
        if self.aai_indices != None:
            self.aai_indices = None

        #print out results from encoding
        self.output_results(desc_df)

        #plot regression plot for predictive model
        plot_reg(Y_test, Y_pred, eval.r2, show_plot)

        #save results of encoding to output folder
        save_results(desc_df, 'desc_encoding')

        #reset descriptors instance variable
        self.descriptors = None #**

        return desc_df

    def encode_aai_desc(self, indices=None, descriptors=None, show_plot=False):
        """
        Encode using both AAI indices and the descriptors. The two outputs from
        the individual encoding strategies will be concatenated together and
        used in the building of a predictive model. The resulting model assets
        and its results will be exported to the directory pointed to by the global
        variable OUTPUT_DIR. If the config parameter use_dsp is true then pass 
        AAI Indices through a DSP transformation specified by DSP parameters 
        (spectrum, window, convolution & filter) via the PyDSP class/module.

        Parameters
        ----------
        :indices : str/list (default=None)
            string or list of indices from the AAI.
        :descriptors : str/list (default=None)
            string or list of protein descriptors.
        :show_plot : bool (default = False)
            display regression plot of best predictive model. If False then the plot
            will just be saved to the output folder, else it'll be displayed & saved.

        Returns
        -------
        :aai_desc_df : pd.Series
            pandas series storing metrics and results of encoding.
        """

        #validate AAI indices and Descriptors are present in the input parameters
        #   or instance variables, return error if either is None
        if (descriptors == None or descriptors == "") or \
            (indices == None or indices == ""):
                raise ValueError('AAI Indices and Descriptor input parameters must \
                    not be empty or None')

        #check input descriptor & indices are of correct type (str/list), if not raise type error
        if (not isinstance(indices, str) and (not isinstance(indices, list)) or \
                (not isinstance(descriptors, str) and (not descriptors(indices, list)))):
            raise TypeError("Input indices and descriptors parameter must be of type string or list.")


        self.aai_indices = indices           #set instance attributes
        self.descriptors = descriptors

        #create output results Series
        aai_desc_df = pd.Series(index=['Index','Category','Descriptor',
            'Group','R2', 'RMSE', 'MSE', 'RPD', 'MAE', 'Explained Variance'], dtype='object')

        #create instance of Descriptors class using data in instance variable
        descr = Descriptors(self.config_file, protein_seqs=self.data[self.sequence_col])

        #get AAI index encoding features
        aai_encoding = self.get_aai_encoding(indices)
        aai_encoding = pd.DataFrame(aai_encoding)

        #if AAI indices encoding is empty, raise error
        if (aai_encoding.empty):
            raise ValueError('AAI Indices {} cannot be empty or None'.format(indices))

        #get descriptor encoding features
        descriptor_encoding = self.get_descriptor_encoding(descriptors)

        #if descriptors encoding is empty, raise error
        if (descriptor_encoding.empty):
            raise ValueError('Descriptors {} cannot be empty or None'.format(descriptors))

        #concatenate AAI index and Descriptor features to get training data (X)
        aai_desc = [aai_encoding, descriptor_encoding]
        X = pd.concat(aai_desc, axis=1)

        #get training and test dataset split
        X_train, X_test, Y_train, Y_test = self.model.train_test_split(X,
            self.get_activity(), test_size=self.test_split)

        #fit predictive model
        self.model.fit()

        #predict activity values for test data
        Y_pred = self.model.predict()

        #create instance of Evaluate class which will get all the evaluation metrics
        eval = Evaluate(Y_test, Y_pred)

        index_cat = ""

        #get categories for all indices in self.aai_indices
        if isinstance(self.aai_indices, list):
            for i in range(0,len(self.aai_indices)):
                index_cat += index_cat + ', ' + \
                    self.aaindex.get_category_from_record(self.aai_indices[i])
        else:
            index_cat = self.aaindex.get_category_from_record(self.aai_indices)

        desc_group = ""
        desc_cat = ""

        #get groups for all descriptors in self.desciptors
        if isinstance(self.descriptors, list):
            for i in range(0,len(self.descriptors)):
                desc_group += desc_group + ', ' + \
                    descr.descriptor_groups[self.descriptors[i]]
        else:
            desc_cat = descr.descriptor_groups[self.descriptors]

        #set output dataframe columns
        aai_desc_df['Index'] = str(self.aai_indices)
        aai_desc_df['Category'] = str(index_cat)
        aai_desc_df['Descriptor'] = str(self.descriptors)
        aai_desc_df['Group'] = str(desc_cat)
        aai_desc_df['R2'] = eval.r2
        aai_desc_df['RMSE'] = eval.rmse
        aai_desc_df['MSE'] = eval.mse
        aai_desc_df['RPD'] = eval.rpd
        aai_desc_df['MAE'] = eval.mae
        aai_desc_df['Explained Variance'] = eval.explained_var

        #print out results from encoding
        self.output_results(aai_desc_df)

        #plot regression plot for predictive model
        plot_reg(Y_test, Y_pred, eval.r2, show_plot)

        #save results of encoding to output folder
        save_results(aai_desc_df, 'aai_desc_encoding')

        return aai_desc_df

    def output_results(self, results):
        """
        Print out the predictive model parameters/attributes and its results.

        Parameters
        ----------
        :results : dict/pd.Series
            dictionary or Series of metrics and their associated values.
        """
        print('#############################################################')
        print('####################### Parameters ##########################\n')
        print('# Dataset -> {}\n# Dataset Size -> {}\n# Sequence Length -> {} \
            \n# Activity -> {}'.format(self.dataset, self.num_seqs, self.seq_len, self.activity))
        if (self.aai_indices is not None):
            print('# AAI Indices -> {}'.format(self.aai_indices))
            if (self.use_dsp):
                print('  # Using DSP -> {}\n  # Spectrum -> {}\n  # Window Function -> {} \
                \n  # Filter Function -> {}\n  # Convolution -> {}'.format(
                    self.use_dsp, self.spectrum, self.window, self.filter, self.convolution))  
        print('# Descriptors -> {}\n# Algorithm -> {}\
                \n# Model Parameters -> {}\n# Test Split -> {}'.format(
                self.descriptors, repr(self.model), self.model.model.get_params(), self.test_split
                ))
        print('\n#############################################################')
        print('######################## Results ############################')
        # print('\n#############################################################\n')

        # print('######################## Metrics ############################\n')
        print('# R2: {}'.format(results['R2']))
        print('# RMSE: {} '.format(results['RMSE']))
        print('# MSE: {} '.format(results['MSE']))
        print('# MAE: {}'.format(results['MAE']))
        print('# RPD {}'.format(results['RPD']))
        print('# Explained Variance {}\n'.format(results['Explained Variance']))
        print('#############################################################\n')

    def get_seqs(self):
        """
        Return all protein sequences within the dataset.

        Returns
        -------
        :self.data[self.sequence_col] : pd.Series
            pandas series of all protein sequences in the dataset.
        """
        return self.data[self.sequence_col]

    def get_activity(self):
        """
        Return all activity/fitness values for the sequences within the dataset.

        Returns
        -------
        :self.data[self.activity] : np.ndarray
            array of all activity/fitness values for the protein sequences
        """
        return self.data[self.activity].values.reshape((-1,1))

######################          Getters & Setters          ######################

    @property
    def dataset(self):
        return self._dataset

    @dataset.setter
    def dataset(self, val):
        self._dataset = val

    @property
    def sequences(self):
        return self._sequences

    @sequences.setter
    def sequences(self, val):
        self._sequences = val

    @property
    def sequence_col(self):
        return self._sequence_col

    @sequence_col.setter
    def sequence_col(self, val):
        self._sequence_col = val

    @property
    def activity(self):
        return self._activity

    @activity.setter
    def activity(self, val):
        self._activity = val

    # @property
    # def aai_indices(self):
    #     return self._aai_indices

    # @aai_indices.setter
    # def aai_indices(self, val):
    #     self._aai_indices = val

    # @property
    # def descriptors(self):
    #     return self.descriptors

    # @descriptors.setter
    # def descriptors(self, val):
    #     self._descriptors = val

    @property
    def algorithm(self):
        return self._algorithm

    @algorithm.setter
    def algorithm(self, val):
        self._algorithm = val

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, val):
        self._parameters = val

    @property
    def test_split(self):
        return self._test_split

    @test_split.setter
    def test_split(self, val):
        self._test_split = val

    @property
    def num_seqs(self):
        return self._num_seqs

    @num_seqs.setter
    def num_seqs(self, val):
        self._num_seqs = val

    @property
    def seq_len(self):
        return self._seq_len

    @seq_len.setter
    def seq_len(self, val):
        self._seq_len = val

################################################################################

    def __str__(self):
        return "Instance of PySAR class, using parameters: {}".format(self.__dict__)

    def __repr__(self):
        return "<PySAR: {}>".format(self)



import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)
import unittest

from pySAR.get_protein import *

class GetProteinTests(unittest.TestCase):
    """
    Unit tests for testing the get_protein module of pySAR.
    """
    def test_fasta(self):
        """ Test get_protein_from_fasta function which returns protein seq from fasta file. """
        test_seq = get_protein_from_fasta(os.path.join("tests", "test_data", "test_fasta.fasta"))
#1.)
        with self.assertRaises(OSError):
            test_seq = get_protein_from_fasta("incorrect_filepath.fasta")
        with self.assertRaises(ValueError):
            test_seq.index("Z")
        with self.assertRaises(ValueError):
            test_seq.index("B")
#2.)
        self.assertEqual(len(test_seq), 91)
        self.assertTrue(test_seq.startswith('MNVKKA'))
#3.)
        self.assertEqual(test_seq.find("X"), -1)
        self.assertEqual(test_seq.find("B"), -1)
        self.assertEqual(test_seq.find("Z"), -1)

    @unittest.skip("Don't want to overload the Uniprot server each time tests are run.")
    def test_protein_uniprot(self):
        """ Testing getting protein sequence from Uniprot database. """
        protid_1 = "P94485"
        protid_2 = "Q7CPU9"
        protid_3 = "Q802Y8"
        protid_4 = "Q04731"
        invalid_protid_1 = "abcde"
        invalid_protid_2 = "fghi"
        protid_savedir = os.path.join("tests", "test_data")

        download_protein_from_uniprot(protid_1, save_dir=protid_savedir)
        download_protein_from_uniprot(protid_2, save_dir=protid_savedir)
        download_protein_from_uniprot(protid_3, save_dir=protid_savedir)
        download_protein_from_uniprot(protid_4, save_dir=protid_savedir)
#1.)
        self.assertTrue(os.path.isfile(os.path.join(protid_savedir, protid_1)))
        self.assertTrue(os.path.isfile(os.path.join(protid_savedir, protid_2)))
        self.assertTrue(os.path.isfile(os.path.join(protid_savedir, protid_3)))
        self.assertTrue(os.path.isfile(os.path.join(protid_savedir, protid_4)))
#2.)
        with self.assertRaises(requests.exceptions.RequestException, msg='Requests Error raised, invalid url given.'):
            download_protein_from_uniprot(invalid_protid_1, save_dir=protid_savedir)

        with self.assertRaises(requests.exceptions.RequestException, msg='Requests Error raised, invalid url given.'):
            download_protein_from_uniprot(invalid_protid_2, save_dir=protid_savedir)

    @unittest.skip("Don't want to overload the NCBI server each time tests are run.")
    def test_protein_ncbi(self):
        """ Testing getting protein sequence from NCBI database. """
        ncbi_1 = "P94485"
        ncbi_2 = "Q7CPU9"
        ncbi_3 = "Q802Y8"
        ncbi_4 = "Q04731"
        invalid_protid_1 = "abcde"
        invalid_protid_2 = "fghi"
        ncib_savedir = os.path.join("tests", "test_data")

        download_protein_from_ncbi(ncbi_1, save_dir=ncib_savedir)
        download_protein_from_ncbi(ncbi_2, save_dir=ncib_savedir)
        download_protein_from_ncbi(ncbi_3, save_dir=ncib_savedir)
        download_protein_from_ncbi(ncbi_4, save_dir=ncib_savedir)
#1.)
        self.assertTrue(os.path.isfile(os.path.join('pySAR', ncib_savedir, ncbi_1)))
        self.assertTrue(os.path.isfile(os.path.join('pySAR', ncib_savedir, ncbi_2)))
        self.assertTrue(os.path.isfile(os.path.join('pySAR', ncib_savedir, ncbi_3)))
        self.assertTrue(os.path.isfile(os.path.join('pySAR', ncib_savedir, ncbi_4)))
#2.)
        with self.assertRaises(requests.exceptions.RequestException, msg='Requests Error raised, invalid url given.'):
            download_protein_from_ncbi(invalid_protid_1, save_dir=ncib_savedir)

        with self.assertRaises(requests.exceptions.RequestException, msg='Requests Error raised, invalid url given.'):
            download_protein_from_ncbi(invalid_protid_2, save_dir=ncib_savedir)

    @unittest.skip("Don't want to overload the PDB server each time tests are run.")
    def test_protein_pdb(self):
        """ Testing getting protein sequence from PDB database. """
        pdb_1 = "7MLP"
        pdb_2 = "2GPZ"
        pdb_3 = "7JPH"
        pdb_4 = "7JPI"
        invalid_protid_1 = "abcde"
        invalid_protid_2 = "fghi"
        pdb_savedir = os.path.join("tests", "test_data")

        download_protein_from_pdb(pdb_1, save_dir=pdb_savedir)
        download_protein_from_pdb(pdb_2, save_dir=pdb_savedir)
        download_protein_from_pdb(pdb_3, save_dir=pdb_savedir)
        download_protein_from_pdb(pdb_4, save_dir=pdb_savedir)
#1.)
        self.assertTrue(os.path.isfile(os.path.join('pySAR', pdb_savedir, pdb_1)))
        self.assertTrue(os.path.isfile(os.path.join('pySAR', pdb_savedir, pdb_2)))
        self.assertTrue(os.path.isfile(os.path.join('pySAR', pdb_savedir, pdb_3)))
        self.assertTrue(os.path.isfile(os.path.join('pySAR', pdb_savedir, pdb_4)))
#2.)
        with self.assertRaises(requests.exceptions.RequestException, msg='Requests Error raised, invalid url given.'):
            download_protein_from_ncbi(invalid_protid_1, save_dir=pdb_savedir)

        with self.assertRaises(requests.exceptions.RequestException, msg='Requests Error raised, invalid url given.'):
            download_protein_from_ncbi(invalid_protid_2, save_dir=pdb_savedir)


################################################################################
#################                    Evaluate                  #################
################################################################################

import numpy as np
import pandas as pd
from sklearn.metrics import r2_score, mean_squared_error, mean_absolute_error, \
    explained_variance_score, max_error, mean_poisson_deviance

class Evaluate():
    """
    An instance of the Evaluate class will calculate various metric values for
    the inputted observed (Y_true) and predicted (Y_pred) arrays, storing the
    results in the class attributes. The class supports metrics: R2, RMSE, MSE,
    MAE, RPD, Explained Variance, Max Error and Mean Poisson Deviance.

    Parameters
    ----------
    :Y_true : np.ndarray
        array of observed activity/fitness values.
    :Y_pred : np.ndarray
        array of predicted activity/fitness values.

    Methods
    -------
    r2_(multioutput='uniform_average'):
        calculate R2 score.
    rmse_(multioutput='uniform_average'):
        calculate RMSE value
    mse_(multioutput='uniform_average'):
        calculate MSE value.
    mae_(multioutput='uniform_average'):
        calculate MAE value.
    rpd_():
        calculate ratio of performance to deviation.
    explained_var_(multioutput='uniform_average'):
        calculate explained variance.
    max_error_():
        calculate max errror.
    mean_poisson_deviance_():
        calculate mean poisson deviance.
    """
    def __init__(self, Y_true, Y_pred):

        #convert input observed and predicted values into numpy arrays and reshape
        self.Y_true = np.array(Y_true).reshape((-1,1))
        self.Y_pred = np.array(Y_pred).reshape((-1,1))

        #validate that predicted and observed input arrays are of the same length,
        #   if input predicted and observed arrays are not same shape then raise error
        if (self.Y_true.shape) != (self.Y_pred.shape):
            raise ValueError('Observed and predicted values must be of the same length, \
                Y_true = {} & Y_pred = {}.'.format(Y_true.shape, Y_pred.shape))

        #calculate all metric values for inputs
        self.r2 = self.r2_()
        self.rmse = self.rmse_()
        self.mse = self.mse_()
        self.mae = self.mae_()
        self.rpd = self.rpd_()
        self.explained_var = self.explained_var_()
        # self.mean_poisson_deviance = self.mean_poisson_deviance_()
        self.max_error = self.max_error_()

    def r2_(self, multioutput='uniform_average'):
        """
        Calculate R^2 (coefficient of determination) regression score function.

        Parameters
        ----------
        :multioutput : str (default='uniform_average')
            method that defines aggregating of multiple output scores. Default
            is reccomended ('uniform_average'), available values:
            {‘raw_values’, ‘uniform_average’, ‘variance_weighted’}.
        Returns
        -------
        :r2 : float
            R2 (coefficient of determination) score for observed and predicted values.
        """
        return r2_score(self.Y_true, self.Y_pred, multioutput=multioutput)

    def mse_(self,multioutput='uniform_average'):
        """
        Calculate MSE (mean square error) regression loss score for observed
        and predicted values.

        Parameters
        ----------
        :multioutput : str (default = 'uniform_average')
            method that defines aggregating of multiple output scores. Default
            is reccomended ('uniform_average'), available values:
            {‘raw_values’, ‘uniform_average’, ‘variance_weighted’}.
        Returns
        -------
        :mse : float
            MSE (mean square error) score for observed and predicted values.
        """
        return mean_squared_error(self.Y_true, self.Y_pred, multioutput=multioutput)

    def rmse_(self, multioutput='uniform_average'):
        """
        Calculate the RMSE (root mean square error) regression loss score for
        inputted observed and predicted values. Uses the same function for
        calculating MSE with the squared parameter set to False.

        Parameters
        ----------
        :multioutput : str (default='uniform_average')
            method that defines aggregating of multiple output scores. Default
            is reccomended ('uniform_average'), available values:
            {‘raw_values’, ‘uniform_average’, ‘variance_weighted’}.
        Returns
        -------
        :rmse : float
            RMSE score for observed and predicted values.
        """
        return mean_squared_error(self.Y_true, self.Y_pred, squared=False, multioutput=multioutput)

    def mae_(self, multioutput='uniform_average'):
        """
        Calculate the Mean Absolute Error (MAE) regression loss for
        inputted observed and predicted values.

        Parameters
        ----------
        :multioutput : str (default='uniform_average')
            method that defines aggregating of multiple output scores. Default
            is reccomended ('uniform_average'), available values:
            {‘raw_values’, ‘uniform_average’, ‘variance_weighted’}.

        Returns
        -------
        :mae : float
            If multioutput is ‘raw_values’, then MAE is returned for each output
            separately. If multioutput is ‘uniform_average’ or an ndarray of
            weights, then the weighted average of all output errors is returned.
            The output is a non-negative floating point. The best value is 0.0.
        """
        return mean_absolute_error(self.Y_true, self.Y_pred, multioutput=multioutput)

    def rpd_(self):
        """
        Calculates the Ratio of Performance to Deviation (RPD). RPD is the ratio
        between the standard deviation of a variable and the standard error of
        prediction of that variable by a given model.

        Returns
        -------
        :rpd : float
            the RPD score for the model.
        """
        return self.Y_true.std() / np.sqrt(self.mse_())

    def explained_var_(self, multioutput='uniform_average'):
        """
        Calculates the Explained Variance regression score. Best possible score is 1.0,
        lower values are worse.

        Parameters
        ----------
        :multioutput : str (default='uniform_average')
            method that defines aggregating of multiple output scores. Default
            is reccomended ('uniform_average'), available values:
            {‘raw_values’, ‘uniform_average’, ‘variance_weighted’}.

        Returns
        -------
        :explained_var : float
            The explained variance or ndarray if ‘multioutput’ is ‘raw_values’.
        """
        return explained_variance_score(self.Y_true, self.Y_pred, multioutput=multioutput)

    def max_error_(self):
        """
        Calculates the maximum residual error between observed and predicted
        values.

        Returns
        -------
        :max_error : float
            A positive floating point value (the best value is 0.0).
        """
        return float(max_error(self.Y_true, self.Y_pred))

    def mean_poisson_deviance_(self):
        """
        Calculate Mean Poisson deviance regression loss.

        Returns
        -------
        :mean_poisson_deviance : float
            A non-negative floating point value (the best value is 0.0).
        """
        return mean_poisson_deviance(self.Y_true, self.Y_true)

################################################################################

    def __repr__(self):
        return "<Evaluate(Y_true: {} Y_pred: {})>".format(
            self.Y_true.shape, self.Y_pred.shape)

    def __str__(self):
        return "Instance of Evaluate Class with attribute values: \
                R2: {}, RMSE: {}, MSE: {}, MAE: {}, RPD: {}, Explained Variance: {}, \
                    Max Error: {}, Mean Poisson Deviance: {}" \
                .format(self.r2, self.rmse, self.mse, self.mae, self.rpd, self.explained_var,
                    self.max_error, self.mean_poisson_deviance)




################################################################################
#################             Utilities Module Tests           #################
################################################################################
import os
import shutil
import unittest
import numpy as np
# np.warnings.filterwarnings('error', category=np.VisibleDeprecationWarning)
import pandas as pd

import pySAR.globals_ as _globals
import pySAR.utils as utils

class UtilsTest(unittest.TestCase):

    def setUp(self):
        """ Import all test datasets. """
        try:
            self.test_dataset1 = pd.read_csv(os.path.join('tests', 'test_data',
                'test_thermostability.txt'), sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset1.')
        try:
            self.test_dataset2 = pd.read_csv(os.path.join('tests', 'test_data',
                'test_enantioselectivity.txt'), sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset2.')
        try:
            self.test_dataset3 = pd.read_csv(os.path.join('tests', 'test_data',
                'test_localization.txt'), sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset3.')
        try:
            self.test_dataset4 = pd.read_csv(os.path.join('tests', 'test_data',
                'test_absorption.txt'), sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset4.')

        #append all datasets to a list
        self.all_test_datasets = [self.test_dataset1, self.test_dataset2, 
                self.test_dataset3, self.test_dataset4]

        #set global vars to create temp test data folders
        _globals.OUTPUT_DIR = os.path.join('tests', _globals.OUTPUT_DIR)
        _globals.OUTPUT_FOLDER = os.path.join('tests', _globals.OUTPUT_FOLDER)

    def test_valid_sequence(self):
        """ Test Valid/Invalid Sequences utility function. """
        invalid_seqs = [["A", "B", "C", "D"],["E","F","J"]]
        invalid_seqs1 = ["ZZZZZZ"]
        invalid_seqs2 = [["Z", 2, "Y", "X", 321]]
        invalid_seqs3 = "XXZXXZXXZ"
#1.)
        #testing with invalid sequences
        self.assertIsNotNone(utils.valid_sequence(invalid_seqs))
        self.assertIsNotNone(utils.valid_sequence(invalid_seqs1))
        self.assertIsNotNone(utils.valid_sequence(invalid_seqs2))
        self.assertIsNotNone(utils.valid_sequence(invalid_seqs3))
#2.)
        self.assertEqual(len(utils.valid_sequence(invalid_seqs)), 2)
        self.assertEqual(len(utils.valid_sequence(invalid_seqs1)), 6)
        self.assertEqual(len(utils.valid_sequence(invalid_seqs2)), 4)
        self.assertEqual(len(utils.valid_sequence(invalid_seqs3)), 9)
#3.)
        self.assertIsInstance((utils.valid_sequence(invalid_seqs)), list)
        self.assertIsInstance((utils.valid_sequence(invalid_seqs1)), list)
        self.assertIsInstance((utils.valid_sequence(invalid_seqs2)), list)
        self.assertIsInstance((utils.valid_sequence(invalid_seqs3)), list)
#4.)
        #testing with valid sequences
        random_seq = np.random.randint(0, len(self.test_dataset1))
        random_seq1 = np.random.randint(0, len(self.test_dataset1))
        random_seq2 = np.random.randint(0, len(self.test_dataset1))

        self.assertIsNone(utils.valid_sequence(self.test_dataset1['sequence'][random_seq]))
        self.assertIsNone(utils.valid_sequence(self.test_dataset1['sequence'][random_seq1]))
        self.assertIsNone(utils.valid_sequence(self.test_dataset1['sequence'][random_seq2]))

    def test_remove_gaps(self):
        """ Test utility function that removes any gaps from sequences. """
        seq1 = ["A", "B", "C", "D", "-"]
        seq2 = ["A", "B", "C", "D", "-", "-", "-", "E", "F", "-"]
        seq3 = 'ABCDFSDJWD---'
        seq4 = "YUJBVFGHYJ---ASD"
#1.)
        seq1_test = utils.remove_gaps(seq1)
        self.assertEqual(len(seq1_test), 1)
        self.assertEqual(len(seq1_test[0]), 4)
        self.assertIsInstance(seq1_test, list)
        self.assertNotIn('-', seq1_test)
#2.)
        seq2_test = utils.remove_gaps(seq2)
        self.assertEqual(len(seq2_test), 1)
        self.assertEqual(len(seq2_test[0]), 6)
        self.assertIsInstance(seq2_test, list)
        self.assertNotIn('-', seq2_test)
#3.)
        seq3_test = utils.remove_gaps(seq3)
        self.assertEqual(len(seq3_test), 10)
        self.assertIsInstance(seq3_test, str)
        self.assertNotIn('-', seq3_test)
#4.)
        seq4_test = utils.remove_gaps(seq4)
        self.assertEqual(len(seq4_test), 13)
        self.assertIsInstance(seq4_test, str)
        self.assertNotIn('-', seq4_test)

    def test_flatten(self):
        """ Test flatten utility function that flattens an array or list. """
        seq1 = np.array([[1, 2, 3], [4, 5, 6]], np.int32)
        seq2 = np.array([[1, 2, 3], [4, 5, 6],[7, 8, 9]], np.int32)
        seq3 = np.random.randint(10,90,(4,5,2))
        seq4 = ["A", "B", "C", "D", "E", "F"]
        seq5 = "TUVWXYZ"
#1.)
        flattened_array = utils.flatten(seq1)
        self.assertEqual(flattened_array.shape, (6,1))
        self.assertIsInstance(flattened_array, np.ndarray)
        self.assertEqual(flattened_array.ndim, 2)
        self.assertTrue((np.array([[1],[2],[3],[4],[5],[6]]) == flattened_array).all())
#2.)
        flattened_array_2 = utils.flatten(seq2)
        self.assertEqual(flattened_array_2.shape, (9,1))
        self.assertIsInstance(flattened_array_2, np.ndarray)
        self.assertEqual(flattened_array_2.ndim, 2)
        self.assertTrue((np.array([[1],[2],[3],[4],[5],[6],[7],[8],[9]]) == flattened_array_2).all())
#3.)
        flattened_array_3 = utils.flatten(seq3)
        self.assertEqual(flattened_array_3.shape, (40,1))
        self.assertIsInstance(flattened_array_3, np.ndarray)
        self.assertEqual(flattened_array_3.ndim, 2)
#4.)
        flattened_array_4 = utils.flatten(seq4)
        self.assertEqual(len(flattened_array_4), 6)
        self.assertIsInstance(flattened_array_4, list)
        self.assertEqual(flattened_array_4,seq4)
#5.)
        flattened_array_5 = utils.flatten(seq5)
        self.assertEqual(flattened_array_5, seq5)
        self.assertIsInstance(flattened_array_5, str)

    def test_zero_padding(self):
        """ Test zero padding utility function that pads an array or list with 0's. """
        seq1 = np.array([[1, 2, 3, 4, 5], [6, 7, 8]], dtype=object)
        seq2 = [["A", "B", "C"], ["D", "E", "F", "G"],["H"]]
        seq3 = np.random.randint(10,90,(4,5,2))
#1.)
        test_dataset3_padded = utils.zero_padding((self.test_dataset3['sequence']))

        #verify all sequences in dataset have been zero-padded to same length
        for seq in range(0, len(test_dataset3_padded)):
            self.assertEqual(len(test_dataset3_padded[seq]), 361)
            self.assertIsInstance(test_dataset3_padded[seq], str)

        self.assertIsInstance(test_dataset3_padded, pd.Series)
        self.assertEqual(test_dataset3_padded.shape[0], 254)
#2.)
        padded_seqs1 = utils.zero_padding(seq1)
        self.assertEqual(len(padded_seqs1), 2)
        self.assertEqual(len(padded_seqs1[0]), 5)
        self.assertIsInstance(padded_seqs1, np.ndarray)
#3.)
        padded_seqs2 = utils.zero_padding(seq2)
        self.assertEqual(len(padded_seqs2), 3)
        self.assertEqual(len(padded_seqs2[0]), 4)
        self.assertIsInstance(padded_seqs2, list)
#4.)
        padded_seqs3 = utils.zero_padding(seq3)
        self.assertEqual(len(padded_seqs3), 4)
        self.assertEqual(padded_seqs3.shape, (4,5,2))
        self.assertIsInstance(padded_seqs3, np.ndarray)
        self.assertTrue(padded_seqs3.any() == seq3.any())

    @unittest.skip('')
    def test_create_output_dir(self):
        """ Testing create output directory utility function. """
        utils.create_output_dir()
#1.)
        #verify folders and directories have been created
        self.assertTrue(os.path.isdir(_globals.OUTPUT_DIR))
        self.assertTrue(os.path.isdir(_globals.OUTPUT_FOLDER))

    @unittest.skip('')
    def test_save_results(self):
        """ Testing save results utility function. """
        #create output dir to save results
        utils.create_output_dir()
#1.)
        #create dummy test results, save to csv and verify csv has been created & saved
        test_results = {'R2': 0.56, 'MSE': 0.34, 'RMSE': 0.89}
        utils.save_results(test_results, 'test_results')
        self.assertTrue(os.path.isfile(os.path.join(_globals.OUTPUT_FOLDER, 'test_results.csv')))
#2.)
        test_results1 = {'MAE': 2.10, 'MSE': 0.99, 'RPD': 1.28}
        utils.save_results(test_results1, 'test_results1')
        self.assertTrue(os.path.isfile(os.path.join(_globals.OUTPUT_FOLDER, 'test_results1.csv')))
#3.)
        test_results2 = pd.DataFrame(np.random.randint(1,100, size=(5,3)), columns=['R2','MSE','RMSE'])
        utils.save_results(test_results2, 'test_results2')
        self.assertTrue(os.path.isfile(os.path.join(_globals.OUTPUT_FOLDER, 'test_results2.csv')))
#4.)
        test_results3 = pd.Series(np.random.randint(1,100), index=['Col1','Col2','Col3','Col4'])
        utils.save_results(test_results3, 'test_results3')
        self.assertTrue(os.path.isfile(os.path.join(_globals.OUTPUT_FOLDER, 'test_results3.csv')))
#5.)
        test_results4 = np.random.randint(1,100, size=(2,6))
        with self.assertRaises(TypeError, msg='Type Error raised, invalid input parameter data type given.'):
            utils.save_results(test_results4, 'test_results4')
        self.assertFalse(os.path.isfile(os.path.join(_globals.OUTPUT_FOLDER, 'test_results4.csv')))

    def tearDown(self):
        """ Delete any temp files or folders created during testing process. """
        del self.test_dataset1
        del self.test_dataset2
        del self.test_dataset3
        del self.test_dataset4

        #removing any of the temp files created such as the results files, if
        #   you want to verify the results files are actually being created then
        #       comment out the below code block.
        if (os.path.isdir(_globals.OUTPUT_DIR)):
            shutil.rmtree(_globals.OUTPUT_DIR, ignore_errors=False, onerror=None)
        if (os.path.isdir(_globals.OUTPUT_FOLDER)):
            shutil.rmtree(_globals.OUTPUT_FOLDER, ignore_errors=False, onerror=None)



################################################################################
#################                    Plots                     #################
################################################################################

import matplotlib.pyplot as plt
import seaborn as sns
import os

from .globals_ import OUTPUT_DIR, OUTPUT_FOLDER, DATA_DIR

def plot_reg(Y_true, Y_pred, r2, show_plot=False):
    """
    Plot regression plot of observed (Y_true) vs predicted activity values (Y_pred).

    Parameters
    ----------
    :Y_true : np.ndarray
        array of observed values.
    :Y_pred : np.ndarray
        array of predicted values
    :r2 : float
        r2 score value
    :show_plot : bool (default = False)
        whether to display plot or not when function is run, if False the plot is just
        saved to output folder. 

    Returns
    -------
    None
    """
    fig, ax = plt.subplots(figsize=(8, 8))
    ax = sns.regplot(x=Y_true,y=Y_pred,  marker="+", truncate=False,fit_reg=True)
    r2_annotation = 'R2: {:.3f} '.format(r2)
    ax.text(0.15, 0.92, r2_annotation, ha="left", va="top", fontsize=15, color="green",
        fontweight="bold",transform=ax.transAxes)
    plt.xlabel('Predicted Value',fontdict=dict(weight='bold'), fontsize=12)
    plt.ylabel('Observed Value',fontdict=dict(weight='bold'), fontsize=12)
    plt.title('Observed vs Predicted values for protein activity',fontdict=dict(weight='bold'), fontsize=15)
    plt.savefig(os.path.join(OUTPUT_FOLDER,'model_regPlot.png'))  #save plot to output folder
    if (show_plot):     
        plt.show(block=False)
        plt.pause(3)
        plt.close()



################################################################################
##############                 Global Variables                  ###############
################################################################################

import os
from datetime import datetime

#data dir is the default directory used to store all data required for pySAR
global DATA_DIR
DATA_DIR = 'data'

#output dir is the default directory used to store all outputs generated
global OUTPUT_DIR
OUTPUT_DIR = 'outputs'

#current datetime appended to output assets & directories to uniquely identify them
global CURRENT_DATETIME
CURRENT_DATETIME = str(datetime.date(datetime.now())) + \
    '_' + str((datetime.now().strftime('%H_%M')))

#output folder is the default folder within the OUTPUT_DIR used to store all
#   outputs generated from one run of the program.
global OUTPUT_FOLDER
OUTPUT_FOLDER = os.path.join(OUTPUT_DIR, 'model_output_' + CURRENT_DATETIME)




################################################################################
#################             pySAR Module Tests             #################
################################################################################

import pandas as pd
import numpy as np
import os
import shutil
import unittest
unittest.TestLoader.sortTestMethodsUsing = None

import pySAR.pySAR as pysar
import pySAR.globals_ as _globals

class PySARTests(unittest.TestCase):

    def setUp(self):
        """ Import the 4 config files for each of the 4 datasets used for testing the pySAR methods. """
        #array of config files for each test dataset
        config_path = os.path.join('tests', 'test_config')
        self.all_config_files = [
            os.path.join(config_path, "test_thermostability.json"), 
            os.path.join(config_path, "test_enantioselectivity.json"),
            os.path.join(config_path, "test_absorption.json"), 
            os.path.join(config_path, "test_localization.json")
        ]

        #set global vars to create temp test data folders
        _globals.OUTPUT_DIR = os.path.join('tests', _globals.OUTPUT_DIR)
        _globals.OUTPUT_FOLDER = os.path.join('tests', _globals.OUTPUT_FOLDER)

    def test_pySAR(self):
        """ Testing pySAR intialisation process and associated methods & attributes. """
        test_pySAR = pysar.PySAR(config_file=self.all_config_files[0])
#1.)
        #testing attribute values, including default values
        self.assertEqual(test_pySAR.dataset, (os.path.join('tests', 'test_data', 'test_thermostability.txt')),
            'Dataset attribute does not equal what was input, got {}.'.format(test_pySAR.dataset))
        self.assertEqual(test_pySAR.sequence_col, "sequence",
            'Sequence column attribute is not correct, got {}, expected {}.'.format(test_pySAR.sequence_col, "sequence"))
        self.assertEqual(test_pySAR.activity, "T50",
            "Activity attribute name not correct, expected {}, got {}.".format("T50", test_pySAR.activity))
        self.assertEqual(test_pySAR.algorithm, "PLSRegression",
            'Algorithm attribute not correct, expected {}, got {}.'.format("PLSRegression", test_pySAR.algorithm))
        self.assertEqual(test_pySAR.test_split, 0.2,
            'Test split not expected, got {}, expected {}.'.format(test_pySAR.test_split, 0.2))
        self.assertIsNone(test_pySAR.aai_indices)
        self.assertIsNone(test_pySAR.descriptors)
        self.assertEqual(test_pySAR._parameters, {},
            'Parameters attribute expected to be empty, got {}.'.format(test_pySAR.parameters))
        self.assertEqual(str(type(test_pySAR.aaindex)), "<class 'pySAR.aaindex.AAIndex'>",
            'AAIndex expected to be an instance of the AAIndex class, got {}.'.format(type(test_pySAR.aaindex)))
        self.assertIsInstance(test_pySAR.data, pd.DataFrame,
            'Data expected to be a DataFrame, got {}.'.format(type(test_pySAR.data)))
        self.assertEqual(test_pySAR.data.isnull().sum().sum(), 0,
            'NAN/null values found in data dataframe')
        self.assertEqual(test_pySAR.num_seqs, 261,
            'Number of sequences expected to be 261, got {}.'.format(test_pySAR.num_seqs))
        self.assertEqual(test_pySAR.seq_len, 466,
            'Sequence length expected to be 466, got {}.'.format(test_pySAR.seq_len))
        self.assertEqual((str(type(test_pySAR.model))), "<class 'pySAR.model.Model'>",
            'Model class attribute expected to be an instance of the Model class, got {}.'.format((str(type(test_pySAR.model)))))
        self.assertEqual(((type(test_pySAR.model.model).__name__)), "PLSRegression",
            'Model type expected to be of type PLSRegression, got {}.'.format((type(test_pySAR.model.model).__name__)))
#2.)
        #validate that if errorneous input parameters are input, that errors are raised
        with self.assertRaises(OSError, msg='OS Error raised, config file not found.'):
            test_pySAR1 = pysar.PySAR(config_file="blahblahblah")

        with self.assertRaises(TypeError, msg='Type Error raised, config file parameter not correct data type.'):
            test_pySAR1 = pysar.PySAR(config_file=101)

    def test_get_seqs(self):
        """ Testing getting the protein sequences from the dataset. """
#1.)
        test_pySAR = pysar.PySAR(config_file=self.all_config_files[0])
        test_seqs = test_pySAR.get_seqs()

        self.assertEqual(test_seqs.shape, (test_pySAR._num_seqs, ),
            'Shape of the sequences not correct, expected {}, got {}.'.format(test_seqs.shape, (test_pySAR._num_seqs, )))
        self.assertIsInstance(test_seqs, pd.Series,
            'Sequences not correct type, expected {}, got {}.'.format(pd.Series, type(test_seqs)))
        self.assertTrue(test_seqs[0].startswith("MTIKEMPQPK"),
            'Error in first seqeuence, expected it to start with MTIKEMPQPK.')
        self.assertEqual(test_seqs.dtype, object,
            'Sequence object expected to be of dtype object, got {}.'.format(test_seqs.dtype))
#2.)
        test_pySAR_2 = pysar.PySAR(config_file=self.all_config_files[1])
        test_seqs = test_pySAR_2.get_seqs()

        self.assertEqual(test_seqs.shape, (test_pySAR_2._num_seqs, ),
            'Shape of the sequences not correct, expected {}, got {}.'.format(test_seqs.shape, (test_pySAR_2._num_seqs, )))
        self.assertIsInstance(test_seqs, pd.Series,
            'Sequences not correct type, expected {}, got {}'.format(pd.Series, type(test_seqs)))
        self.assertTrue(test_seqs[0].startswith("MSAPFAKF"),
            'Error in second seqeuence expected it to start with MSAPFAKF.')
        self.assertEqual(test_seqs.dtype, object,
            'Sequence object expected to be of dtype object, got {}.'.format(test_seqs.dtype))
#3.)
        test_pySAR_3 = pysar.PySAR(config_file=self.all_config_files[2])
        test_seqs = test_pySAR_3.get_seqs()

        self.assertEqual(test_seqs.shape, (test_pySAR_3._num_seqs, ),
            'Shape of the sequences not correct, expected {}, got {}.'.format(test_seqs.shape, (test_pySAR_3._num_seqs, )))
        self.assertIsInstance(test_seqs, pd.Series,
            'Sequences not correct type, expected {}, got {}'.format(pd.Series, type(test_seqs)))
        self.assertTrue(test_seqs[0].startswith("MLMTVFSSAP"),
            'Error in third seqeuence expected it to start with MLMTVFSSAP.')
        self.assertEqual(test_seqs.dtype, object,
            'Sequence object expected to be of dtype object, got {}'.format(test_seqs.dtype))
#4.)
        test_pySAR_4 = pysar.PySAR(config_file=self.all_config_files[3])
        test_seqs = test_pySAR_4.get_seqs()

        self.assertEqual(test_seqs.shape, (test_pySAR_4._num_seqs, ),
            'Shape of the sequences not correct, expected {}, got {}.'.format(test_seqs.shape, (test_pySAR_4._num_seqs, )))
        self.assertIsInstance(test_seqs, pd.Series,
            'Sequences not correct type, expected {}, got {}'.format(pd.Series, type(test_seqs)))
        self.assertTrue(test_seqs[0].startswith("MSRLVAASWL"),
            'Error in third seqeuence expected it to start with MSRLVAASWL.')
        self.assertEqual(test_seqs.dtype, object,
            'Sequence object expected to be of dtype object, got {}'.format(test_seqs.dtype))

    def test_get_activity(self):
        """ Testing function that gets activity from dataset. """
#1.)
        test_pySAR = pysar.PySAR(config_file=self.all_config_files[0])
        activity = test_pySAR.get_activity()

        self.assertEqual(activity.shape[0], test_pySAR.num_seqs,
            'Shape of 1st dimension of Activity expected to be {} , got {} '.format(test_pySAR.num_seqs, activity.shape[0]))
        self.assertEqual(activity.shape[1], 1,
            'Shape of 2nd dimension of Activity expected to be 1, got {} '.format(activity.shape[1]))
        self.assertIsInstance(activity, np.ndarray,
            'Activity attribute should of type numpy array, got {}'.format(type(activity)))
        self.assertTrue((activity[:10] == np.array(([55. ],[43. ],[49. ],[39.8],\
            [52.9],[48.8],[45. ],[48.3],[61.5],[54.6]))).all())
#2.)
        test_pySAR_2 = pysar.PySAR(config_file=self.all_config_files[1])
        activity_2 = test_pySAR_2.get_activity()

        self.assertEqual(activity_2.shape[0], test_pySAR_2.num_seqs,
            'Shape of 1st dimension of Activity should be {} , got {} '.format(test_pySAR_2.num_seqs, activity_2.shape[0]))
        self.assertEqual(activity_2.shape[1], 1,
            'Shape of 2nd dimension of Activity should be 1, got {} '.format(activity.shape[1]))
        self.assertIsInstance(activity_2, np.ndarray,
            'Activity attribute should of type numpy array, got {}'.format(type(activity_2)))
        # self.assertTrue((np.array(activity_2[0][0:3]) == np.array((-4.62693565,-5.59911039,-5.71578825))).all())
#3.)
        test_pySAR_3 = pysar.PySAR(config_file=self.all_config_files[2])
        activity_3 = test_pySAR_3.get_activity()

        self.assertEqual(activity_3.shape[0], test_pySAR_3.num_seqs,
            'Shape of 1st dimension of Activity should be {} , got {} '.format(test_pySAR_3.num_seqs, activity_3.shape[0]))
        self.assertEqual(activity_3.shape[1], 1,
            'Shape of 2nd dimension of Activity should be 1, got {} '.format(activity_3.shape[1]))
        self.assertIsInstance(activity_3, np.ndarray,
            'Activity attribute should of type numpy array, got {}'.format(type(activity_3)))
        # self.assertTrue((activity_3[:10] == np.array(([55. ],[43. ],[49. ],[39.8],\
        #     [52.9],[48.8],[45. ],[48.3],[61.5],[54.6]))).all())
#4.)
        test_pySAR_4 = pysar.PySAR(config_file=self.all_config_files[3])
        activity_3 = test_pySAR_4.get_activity()

        self.assertEqual(activity_3.shape[0], test_pySAR_4.num_seqs,
            'Shape of 1st dimension of Activity should be {} , got {} '.format(test_pySAR_4.num_seqs, activity_3.shape[0]))
        self.assertEqual(activity_3.shape[1], 1,
            'Shape of 2nd dimension of Activity should be 1, got {} '.format(activity_3.shape[1]))
        self.assertIsInstance(activity_3, np.ndarray,
            'Activity attribute should of type numpy array, got {}'.format(type(activity_3)))
        # self.assertTrue((activity_3[:10] == np.array(([55. ],[43. ],[49. ],[39.8],\
        #     [52.9],[48.8],[45. ],[48.3],[61.5],[54.6]))).all())

    def test_get_aai_encoding(self):
        """ Testing getting the AAI encoding from the database for specific indices. """
        aa_indices = ["CHAM810101", "ISOY800103"]
        aa_indices1 = "NAKH920102"
        error_aaindices = ["ABCD1234", "ABCD12345"]
        error_aaindices1 = "XYZ4567"
#1.)
        test_pySAR = pysar.PySAR(config_file=self.all_config_files[0])
        aai_encoding = test_pySAR.get_aai_encoding(aa_indices)

        self.assertIsInstance(aai_encoding, np.ndarray,
            'AAI Encoding output expected to be a numpy array, got datatype {}.'.format(type(aai_encoding)))
        self.assertEqual(aai_encoding.shape[0], test_pySAR.num_seqs,
            'The number of sequences in the dataset expected to be {}, got {}.'.format(test_pySAR.num_seqs, aai_encoding.shape[0]))
        self.assertEqual(aai_encoding.shape[1], test_pySAR.seq_len*len(aa_indices),
            'The length of the sequences expected to be {}, got {}.'.format(test_pySAR.seq_len, str(aai_encoding.shape[1])))
        self.assertEqual(aai_encoding.dtype,np.float32,
            'Datatype of elements in numpy array expected be dtype np.float32, got {}.'.format(aai_encoding.dtype))
        self.assertTrue((np.array([0.78, 0.5, 1.02, 0.68, 0.68, 0.78, 0.36, 0.68, \
            0.36, 0.68], dtype=np.float32)==aai_encoding[0][:10]).all(),
                'The first 10 elements of sequence 0 do not match what was expected.')
#2.)
        test_pySAR_1 = pysar.PySAR(config_file=self.all_config_files[1])
        aai_encoding_1 = test_pySAR_1.get_aai_encoding(aa_indices1)

        self.assertIsInstance(aai_encoding_1, np.ndarray,
            'AAI Encoding output expected to be a numpy array, got datatype {}.'.format(type(aai_encoding_1)))
        self.assertEqual(aai_encoding_1.shape[0], test_pySAR_1.num_seqs,
            'The number of sequences in the dataset expected to be {}, got {}.'.format(test_pySAR_1.num_seqs, aai_encoding_1.shape[0]))
        self.assertEqual(aai_encoding_1.shape[1], test_pySAR_1.seq_len,
            'The length of the sequences expected to be {}, got {}.'.format(test_pySAR_1.seq_len, str(aai_encoding_1.shape[1])))
        self.assertEqual(aai_encoding_1.dtype,np.float32,
            'Datatype of elements in numpy array should be of dtype np.float32, got {}.'.format(aai_encoding_1.dtype))
        # self.assertTrue((np.array([3.79, 3.51, 1.8, 6.11, 9.34, 3.79, 7.21, 4.68, 7.21,  \
        #     6.11],dtype=np.float32)==aai_encoding_1[0][:10]).all(),
        #         'The first 10 elements of sequence 0 do not match what was expected.')

#3.)    #testing errenous indices
        with self.assertRaises(ValueError, msg='ValueError: Errorneous indices have been input.'):
            aai_encoding = test_pySAR_1.get_aai_encoding(error_aaindices)
#4.)
        with self.assertRaises(ValueError, msg='ValueError: Errorneous indices have been input.'):
            aai_encoding = test_pySAR_1.get_aai_encoding(error_aaindices1)
#5.)
        with self.assertRaises(TypeError, msg='TypeError: Errorneous indices datatypes have been input.'):
            aai_encoding = test_pySAR_1.get_aai_encoding(1235)
            aai_encoding = test_pySAR_1.get_aai_encoding(40.89)
            aai_encoding = test_pySAR_1.get_aai_encoding(False)

    def test_aai_encoding(self):
        """ Testing AAI encoding functionality. """
        aa_indices = ["CHAM810101","ISOY800103"]
        aa_indices_1 = "NAKH920102"
        aa_indices_2 = "LIFS790103"
        aa_indices_3 = ["PTIO830101", "QIAN880136", "RACS820110"]
        all_indices = [aa_indices, aa_indices_1, aa_indices_2, aa_indices_3]
        error_aaindices = ["ABCD1234","ABCD12345"]
        error_aaindices1 = "XYZ4567"
        #test with erroneous indices
#1.)
        test_pySAR = pysar.PySAR(config_file=self.all_config_files[0])

        with self.assertRaises(ValueError, msg='ValueError: Indices parameter cannot be None.'):
            test_aai_ = test_pySAR.encode_aai(indices=None)
#2.)
        with self.assertRaises(ValueError, msg='ValueError: Indices parameter cannot be None.'):
            test_aai_ = test_pySAR.encode_aai()
#3.)
        with self.assertRaises(ValueError, msg='ValueError: Erroneous indices put into indices parameter.'):
            test_aai_ = test_pySAR.encode_aai(indices=error_aaindices)
#4.)
        with self.assertRaises(TypeError, msg='TypeError: Indices must be lists or strings.'):
            test_aai = test_pySAR.encode_aai(indices=123)
            test_aai = test_pySAR.encode_aai(indices=0.90)
            test_aai = test_pySAR.encode_aai(indices=False)
            test_aai = test_pySAR.encode_aai(indices=9000)
#5.)
        for index in range(0, len(all_indices)):
            test_aai_ = test_pySAR.encode_aai(indices=all_indices[index])
            self.assertIsInstance(test_aai_, pd.Series, 'Output should be a Series, got {}'.format(type(test_aai_)))
            self.assertEqual(len(test_aai_), 8)
            self.assertEqual(test_aai_.dtype, object)

    # @unittest.skip("Slight error with circular imports, skipping.")
    def test_desc_encoding(self):
        """ Testing Descriptor encoding functionality. """
        desc_1 = "aa_comp"
        desc_2 = "distribution"
        desc_3 = "conjoint_triad"
        desc_4 = ["moranauto", "quasi_seq_order"]
        all_desc = [desc_1, desc_2, desc_3, desc_4]
        error_desc = "blahblahblah"
        error_desc_1 = 123
#1.)
        test_pySAR = pysar.PySAR(config_file=self.all_config_files[0])

        with self.assertRaises(ValueError, msg='ValueError: Descriptor parameter cannot be None.'):
            test_desc = test_pySAR.encode_desc(descriptor=None)

        with self.assertRaises(ValueError, msg='ValueError: Descriptor parameter cannot be None.'):
            test_desc = test_pySAR.encode_desc(descriptor=error_desc)
        
        with self.assertRaises(TypeError, msg='TypeError: Descriptor parameter has to be a strong or list.'):
            test_desc = test_pySAR.encode_desc(descriptor=error_desc_1)
            test_desc = test_pySAR.encode_desc(descriptor=123)
            test_desc = test_pySAR.encode_desc(descriptor=0.90)
            test_desc = test_pySAR.encode_desc(descriptor=False)
            test_desc = test_pySAR.encode_desc(descriptor=9000)
#2.)
        for de in range(0,len(all_desc)):
            test_desc = test_pySAR.encode_desc(descriptor="aa_composition")
            self.assertIsInstance(test_desc,pd.Series, 'Output should be a Series, got {}.'.format(type(test_desc)))
            self.assertEqual(len(test_desc), 8)
            self.assertEqual(test_desc.dtype, object)

    @unittest.skip("Slight error with importing descriptors file, fix.")
    def test_aai_desc_encoding(self):
        """ Testing AAI + Descriptor encoding functionality. """
        aa_indices_1 = "CHAM810101"
        aa_indices_2 = "NAKH920102"
        aa_indices_3 = "LIFS790103"
        aa_indices_4 = ["PTIO830101", "QIAN880136", "RACS820110"]
        desc_1 = "aa_comp"
        desc_2 = "distribution"
        desc_3 = "conjoint_triad"
        desc_4 = ["moranauto", "quasi_seq_order"]

        test_pySAR = pysar.PySAR(config_file=self.all_config_files[0])
#1.)
        with self.assertRaises(ValueError, msg='ValueError: Descriptor and indices parameter cannot both be None.'):
            test_desc = test_pySAR.encode_desc(descriptor=None)
            test_desc = test_pySAR.encode_aai_desc(indices=None)
            test_desc = test_pySAR.encode_aai_desc(descriptor="aa_comp")
            test_desc = test_pySAR.encode_aai_desc(indices="LIFS790103")
#2.)
        with self.assertRaises(ValueError, msg='ValueError: Descriptor and indices must be lists or strings.'):
            test_desc = test_pySAR.encode_desc(descriptor=123)
            test_desc = test_pySAR.encode_aai_desc(indices=0.90)
            test_desc = test_pySAR.encode_aai_desc(descriptor=False)
            test_desc = test_pySAR.encode_aai_desc(indices=9000)
#3.)
        with self.assertRaises(TypeError, msg='TypeError: Descriptor and indices must be lists or strings.'):
            test_aai_desc = test_pySAR.encode_aai_desc(descriptor=123)
            test_aai_desc = test_pySAR.encode_aai_desc(indices=0.90)
            test_aai_desc = test_pySAR.encode_aai_desc(descriptor=False)
            test_aai_desc = test_pySAR.encode_aai_desc(indices=9000)
#4.)
        test_aai_ = test_pySAR.encode_aai_desc(descriptors=desc_1, indices=aa_indices_1, spectrum='power')
        self.assertIsInstance(test_aai_, pd.DataFrame, 'Output should be a DataFrame, got {}'.format(type(test_aai_)))
        self.assertEqual(len(test_aai_), 8)
        self.assertEqual(test_aai_.dtype, object)

    def tearDown(self):
        """ Delete any temp files or folders created during testing process. """
        del self.all_config_files

        #delete any temporary output directory or folders created
        if (os.path.isdir(_globals.OUTPUT_DIR)):
            shutil.rmtree(_globals.OUTPUT_DIR, ignore_errors=False, onerror=None)
        if (os.path.isdir(_globals.OUTPUT_FOLDER)):
            shutil.rmtree(_globals.OUTPUT_FOLDER, ignore_errors=False, onerror=None)

        # del _globals.OUTPUT_DIR
        # del _globals.OUTPUT_FOLDER
                
if __name__ == '__main__':
    #run all unit tests
    unittest.main(verbosity=2)